import os
from graphviz import Digraph
from yattag import Doc, indent
from .types import List
from .types import Path


class HTML:
    """
    This contains helper functions for creating HTML files with yattag.

    Parameters
    ----------
    outFile: str
        Output file
    """

    def __init__(self, outFile: Path):
        """
        This function initializes the header file, and saves useful variables to self.

        Parameters
        ----------
        outFile : Path
            Output file
        """
        self._outFile = outFile
        self.doc, self.tag, self.text = Doc().tagtext()
        self.centerClass = (
            ".center {\n\tdisplay: block;\n\tmargin-left: auto;\n\tmargin-right: auto;\n}\n"
        )

    def GenerateHtml(self) -> str:
        """
        This function generates and formats the HTML file text.

        Returns:
        --------
        html : str
            HTML file as a string.
        """

        html = indent(self.doc.getvalue(), indentation="", newline="\n")
        return html

    def WriteFile(self):
        """This function writes the HTML file text to a file."""
        if not os.path.exists(os.path.dirname(self._outFile)):
            os.makedirs(os.path.dirname(self._outFile))
        out = open(self._outFile, "w")
        out.write(self.GenerateHtml())
        out.close()

    def ReadFile(self, inFile: Path) -> str:
        """This function reads the file specified by "inFile" and retuns the
        contents as a string.

        Parameters
        ----------
        inFile : Path
            File to read.

        Returns
        -------
        outStr : str
            Contents of inFile as a string.
        """
        tmpFile = open(inFile, "r")
        dark = tmpFile.read()
        tmpFile.close()
        return dark


class Dot:
    """
    This class contains helper functions used to create dot graphs.
    """

    def __init__(self, outFile: Path, name: str):
        """
        This function initializes the class and creates the digraph.

        Parameters
        ----------
        outFile : Path
            Name of the filename under which the dot file should be saved.

        name : str
            What the dot file should be called by Digraph.
        """

        self._outFile = outFile
        self._name = name
        self.dot = Digraph(name=self._name)

    def Render(self, formats: List[str] = ["cmapx", "svg"]):
        """
        This function renders the dot graph as a .svg and as a .cmapx.

        Parameters
        ----------
        formats : List[str], optional
            List whose elementts dictate which formats to render the dot graph in. (Default value = ["cmapx", "svg"])
        """
        for f in formats:
            self.dot.render(self._outFile, format=f, cleanup=True, view=False)



from jax._src.config import config

config.update("jax_enable_x64", True)

import numpy as onp
import jax.numpy as np
import numpy.typing as npt
from typing import Optional, cast
from .utils.types import Literal, uint, IntArrayLike, JaxOrNumpyArray
from jax import core
from jax.interpreters import ad, batching, mlir
from jax.lib import xla_client
from jaxlib import hlo_helpers
import jaxlib.mlir.ir as ir

from .utils.TFCUtils import TFCPrint


class utfc:
    """
    This is the univariate TFC class. It acts as a container that creates and stores:

    * The linear map between the free function domain (z) and the problem domain (x).
    * The basis functions or ELMs that make up the free function.
    * The necessary JAX code that enables automatic differentiation of the free function.
    * Other useful TFC related functions such as collocation point creation.

    In addition, this class ties these methods together to form a utility that enables a higher level of code abstraction
    such that the end-user scripts are simple, clear, and elegant implementations of TFC.

    Parameters
    ----------
    N : int
        Number of points to use when discretizing the domain.
    nC : IntArrayLike
        Number of functions to remove from the free function linear expansion. This variable is used to account for basis functions that are linearly dependent on support functions used in the construction of the constrained expressions. It can be expressed in 1 of 2 ways.

        1. As an integer. When expressed as an integer, the first nC basis functions are removed from the free function.
        2. As a list or array. When expressed as a list or array, the basis functions corresponding to the numbers given by the list or array are removed from the free function.

    deg : int
        Degree of the basis function expansion. This number is one less than the number of basis functions used before removing those specified by nC.
    x0 : float, optional
        Specifies the beginning of the DE domain. (Default value = 0)
    xf : float
        This required keyword argument specifies the end of the DE domain.
    basis : Literal["CP","LeP","FS","ELMTanh","ELMSigmoid","ELMSin","ELMSwish","ELMReLU"], optional
        This optional keyword argument specifies the basis functions to be used. (Default value = "CP")
    """

    def __init__(
        self,
        N: uint,
        nC: IntArrayLike,
        deg: uint,
        basis: Literal[
            "CP", "LeP", "FS", "ELMTanh", "ELMSigmoid", "ELMSin", "ELMSwish", "ELMReLU"
        ] = "CP",
        x0: Optional[float] = None,
        xf: Optional[float] = None,
        backend: Literal["C++", "Python"] = "C++",
    ):
        """
        Constructor for the utfc class.

        Parameters
        ----------
        N : int
            Number of points to use when discretizing the domain.
        nC : int or list or array-like
            Number of functions to remove from the free function linear expansion. This variable is used to account for basis functions that are linearly dependent on support functions used in the construction of the constrained expressions. It can be expressed in 1 of 2 ways.
            1. As an integer. When expressed as an integer, the first nC basis functions are removed from the free function.
            2. As a list or array. When expressed as a list or array, the basis functions corresponding to the numbers given by the list or array are removed from the free function.
        deg : int
            Degree of the basis function expansion. This number is one less than the number of basis functions used before removing those specified by nC.
        x0 : float, optional
            Specifies the beginning of the DE domain. (Default value = 0)
        xf : float
            This required keyword argument specifies the end of the DE domain.
        basis : {"CP","LeP","FS","ELMTanh","ELMSigmoid","ELMSin","ELMSwish","ELMReLU"}, optional
            This optional keyword argument specifies the basis functions to be used. (Default value = "CP")
        backend : Literal["C++", "Python"]
            This optional keyword sets the backend used to compute the basis functions. The C++ can be used with JIT, but can only be used for doubles. The Python backend can be used for other field types, e.g., complex numbers, but does not have JIT translations. Instead, pejit must be used to set the basis function outputs as compile time constants in order to JIT.
        """

        # Store givens
        self.N = N
        self.deg = deg
        self._backend = backend

        if isinstance(nC, int):
            self.nC: npt.NDArray = onp.arange(nC, dtype=onp.int32)
        elif isinstance(nC, np.ndarray):
            self.nC: npt.NDArray = cast(npt.NDArray, nC.astype(onp.int32))
        elif isinstance(nC, list):
            self.nC: npt.NDArray = np.array(nC, dtype=np.int32)
        if self.nC.shape[0] > self.deg:
            TFCPrint.Error("Number of basis functions is less than number of constraints!")
        if np.any(self.nC < 0):
            TFCPrint.Error(
                "To set nC to -1 (no constraints) either use nC = -1 or nC = 0 (i.e., use an integer not a list or array). Do not put only -1 in a list or array, this will cause issues in the C++ layer."
            )

        self.basis = basis

        if x0 is None:
            self.x0 = 0.0
        else:
            self.x0 = x0

        if isinstance(self.x0, int):
            self.x0 = float(self.x0)
            TFCPrint.Warning("x0 is an integer. Converting to float to avoid errors down the line.")

        if xf is None:
            self.xf = 0.0
        else:
            self.xf = xf

        if isinstance(self.xf, int):
            self.xf = float(self.xf)
            TFCPrint.Warning("xf is an integer. Converting to float to avoid errors down the line.")

        # Setup the basis function
        if backend == "C++":
            from .utils import BF
        elif backend == "Python":
            from .utils.BF import BF_Py as BF
        else:
            TFCPrint.Error(
                f'The backend {backend} was specified, but can only be one of "C++" or "Python".'
            )
        if self.basis == "CP":
            self.basisClass = BF.CP(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = -1.0
            zf = 1.0
        elif self.basis == "LeP":
            self.basisClass = BF.LeP(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = -1.0
            zf = 1.0
        elif self.basis == "FS":
            self.basisClass = BF.FS(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = -np.pi
            zf = np.pi
        elif self.basis == "ELMReLU":
            self.basisClass = BF.ELMReLU(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        elif self.basis == "ELMSigmoid":
            self.basisClass = BF.ELMSigmoid(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        elif self.basis == "ELMTanh":
            self.basisClass = BF.ELMTanh(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        elif self.basis == "ELMSin":
            self.basisClass = BF.ELMSin(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        elif self.basis == "ELMSwish":
            self.basisClass = BF.ELMSwish(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        else:
            TFCPrint.Error("Invalid basis selection. Please select a valid basis")

        self.c = self.basisClass.c

        # Calculate z points and corresponding x
        if self.basis in ["CP", "LeP"]:
            n = self.N - 1
            # Multiplying x0 by 0 below so the array I has the same
            # type as x0.
            I = np.linspace(0 * self.x0, n, n + 1)
            self.z = np.cos(np.pi * (n - I) / float(n))
            self.x = (self.z - z0) / self.c + self.x0
        else:
            self.z = np.linspace(z0, zf, self.N)
            self.x = (self.z - z0) / self.c + self.x0

        self._SetupJax()

    def H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the basis function matrix for the points specified by x.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        H : NDArray
            Basis function matrix.
        """
        return self._Hjax(x, d=0, full=full)

    def dH(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """This function computes the deriative of H. See documentation of 'H' for more details.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        H : NDArray
            Derivative of the basis function matrix.
        """
        return self._Hjax(x, d=1, full=full)

    def d2H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """This function computes the second deriative of H. See documentation of H for more details.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        d2H : NDArray
            Second derivative of the basis function matrix.
        """
        return self._Hjax(x, d=2, full=full)

    def d4H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """This function computes the fourth deriative of H. See documentation of H for more details.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        d4H : NDArray
            Fourth derivative of the basis function matrix.
        """
        return self._Hjax(x, d=4, full=full)

    def d8H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """This function computes the eighth deriative of H. See documentation of H for more details.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        d8H : NDArray
            Eighth derivative of the basis function matrix.
        """
        return self._Hjax(x, d=8, full=full)

    def _SetupJax(self):
        """This function is used internally by TFC to setup JAX primatives and create desired behavior when taking derivatives of TFC constrained expressions."""

        # Regiser XLA function
        if self._backend == "C++":
            obj = self.basisClass.xlaCapsule
            xlaName_str = "BasisFunc" + str(self.basisClass.identifier)
            xlaName = xlaName_str.encode("utf-8")
            xla_client.register_custom_call_target(xlaName, obj, platform="cpu")

        # Create primitives
        H_p = core.Primitive("H")

        def Hjax(x: JaxOrNumpyArray, d: uint = 0, full: bool = False) -> npt.NDArray:
            return cast(npt.NDArray, H_p.bind(x, d=d, full=full))

        # Implicit translation
        def H_impl(x: npt.NDArray, d: uint = 0, full=False) -> npt.NDArray:
            return self.basisClass.H(x, d, full)

        H_p.def_impl(H_impl)

        # Abstract evaluation
        def H_abstract_eval(x, d: uint = 0, full: bool = False) -> core.ShapedArray:
            if full:
                dim1 = self.basisClass.m
            else:
                dim1 = self.basisClass.m - self.basisClass.numC
            if len(x.shape) == 0:
                dims = (dim1,)
            else:
                dims = (x.shape[0], dim1)
            return core.ShapedArray(dims, x.dtype)

        H_p.def_abstract_eval(H_abstract_eval)

        if self._backend == "C++":
            # XLA compilation
            def default_layout(shape):
                return tuple(range(len(shape) - 1, -1, -1))

            def H_xla(ctx, x, d: uint = 0, full: bool = False):
                x_type = ir.RankedTensorType(x.type)
                dims = x_type.shape
                dim0 = dims[0]
                if full:
                    dim1 = self.basisClass.m
                else:
                    dim1 = self.basisClass.m - self.basisClass.numC
                res_types, res_shapes = hlo_helpers.mk_result_types_and_shapes(
                    [((dim0, dim1), x_type.element_type)]
                )
                return hlo_helpers.custom_call(
                    xlaName_str,
                    result_types=res_types,
                    result_shapes=res_shapes,
                    operands=[
                        hlo_helpers.hlo_s32(self.basisClass.identifier),
                        x,
                        hlo_helpers.hlo_s32(d),
                        mlir.ir_constant(full),
                        hlo_helpers.hlo_s32(dim0),
                        hlo_helpers.hlo_s32(dim1),
                    ],
                    operand_layouts=[(), default_layout(dims), (), (), (), ()],
                    result_layouts=[
                        default_layout((dim0, dim1)),
                    ],
                ).results

            mlir.register_lowering(H_p, H_xla, platform="cpu")

        # Define batching translation
        def H_batch(vec, batch, d: uint = 0, full: bool = False):
            return Hjax(*vec, d=d, full=full), batch[0]

        batching.primitive_batchers[H_p] = H_batch

        # Define jacobain vector product
        def H_jvp(arg_vals, arg_tans, d: uint = 0, full: bool = False):
            x = arg_vals[0]
            dx = arg_tans[0]
            if not (dx is ad.Zero):
                if type(dx) is batching.BatchTracer:
                    flag = onp.any(dx.val != 0)
                else:
                    flag = onp.any(dx != 0)
                if flag:
                    if len(dx.shape) == 1:
                        out_tans = Hjax(x, d=d + 1, full=full) * onp.expand_dims(dx, 1)
                    else:
                        out_tans = Hjax(x, d=d + 1, full=full) * dx
            else:
                dim0 = x.shape[0]
                if full:
                    dim1 = self.basisClass.m
                else:
                    dim1 = self.basisClass.m - self.basisClass.numC
                out_tans = np.zeros((dim0, dim1))
            return (Hjax(x, d=d, full=full), out_tans)

        ad.primitive_jvps[H_p] = H_jvp

        # Provide pointer for TFC class
        self._Hjax = Hjax


class HybridUtfc:
    """
    This class combines TFC classes together so that multiple basis functions can be used
    simultaneously in the solution. Note, that this class is not yet complete.

    Parameters
    ----------
    tfcClasses : list of utfc classes
        This list of utfc classes make up the basis functions used in the HybridUtfc class.
    """

    def __init__(self, tfcClasses):
        """
        This function computes the basis function matrix for the points specified by x.

        Parameters
        ----------
        tfcClasses : list of utfc classes
            This list of utfc classes make up the basis functions used in the HybridUtfc class.
        """

        if not all([k.N == tfcClasses[0].N for k in tfcClasses]):
            TFCPrint.Error("Not all TFC classes provided have the same number of points.")
        self._tfcClasses = tfcClasses

    def H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the basis function matrix for the points specified by x.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        H : NDArray
            Basis function matrix.
        """
        return cast(
            npt.NDArray,
            np.hstack([k._Hjax(x, d=0, full=full) for j, k in enumerate(self._tfcClasses)]),
        )

    def dH(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the derivative of the basis function matrix for the points specified by x.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        dH : NDArray
            Derivative of the basis function matrix.
        """
        return cast(
            npt.NDArray,
            np.hstack([k._Hjax(x, d=1, full=full) for j, k in enumerate(self._tfcClasses)]),
        )

    def d2H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the second derivative of the basis function matrix for the points specified by x.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        d2H : NDArray
            Second derivative of the basis function matrix.
        """
        return cast(
            npt.NDArray,
            np.hstack([k._Hjax(x, d=2, full=full) for j, k in enumerate(self._tfcClasses)]),
        )

    def d3H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the third derivative of the basis function matrix for the points specified by x.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        d3H : NDArray
            Third derivative of the basis function matrix.
        """
        return cast(
            npt.NDArray,
            np.hstack([k._Hjax(x, d=3, full=full) for j, k in enumerate(self._tfcClasses)]),
        )

    def d4H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the fourth derivative of the basis function matrix for the points specified by x.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        d4H : NDArray
            Fourth derivative of the basis function matrix.
        """
        return cast(
            npt.NDArray,
            np.hstack([k._Hjax(x, d=4, full=full) for j, k in enumerate(self._tfcClasses)]),
        )

    def d8H(self, x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the eighth derivative of the basis function matrix for the points specified by x.

        Parameters
        ----------
        x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        d8H : NDArray
            Eighth derivative of the basis function matrix.
        """
        return cast(
            npt.NDArray,
            np.hstack([k._Hjax(x, d=8, full=full) for j, k in enumerate(self._tfcClasses)]),
        )



import numpy as np
import plotly
import plotly.graph_objects as go

from .TFCUtils import TFCPrint
from .types import StrArrayLike, uint, Path, Literal, List
from typing import Optional

TFCPrint()


class MakePlot:
    """
    A MakePlot class for Plotly.
    This class wraps common Plotly functions to ease figure creation.
    """

    _template = {
        "data": {
            "surface": [
                {
                    "colorbar": {"outlinewidth": 0, "ticks": ""},
                    "colorscale": [
                        [0.0, "black"],
                        [0.2, "rebeccapurple"],
                        [0.3, "blueviolet"],
                        [0.7, "#4682B4"],
                        [1.0, "aquamarine"],
                    ],
                    "lighting": {
                        "ambient": 0.6,
                        "specular": 0.05,
                        "diffuse": 0.4,
                        "fresnel": 3.0,
                    },
                    "type": "surface",
                }
            ],
            "volume": [
                {
                    "colorbar": {"outlinewidth": 0, "ticks": ""},
                    "colorscale": [
                        [0.0, "black"],
                        [0.2, "rebeccapurple"],
                        [0.3, "blueviolet"],
                        [0.7, "#4682B4"],
                        [1.0, "aquamarine"],
                    ],
                    "lighting": {
                        "ambient": 0.6,
                        "specular": 0.05,
                        "diffuse": 0.4,
                        "fresnel": 3.0,
                    },
                    "type": "volume",
                }
            ],
            "contour": [
                {
                    "colorbar": {"outlinewidth": 0, "ticks": ""},
                    "colorscale": [
                        [0.0, "black"],
                        [0.2, "rebeccapurple"],
                        [0.3, "blueviolet"],
                        [0.7, "#4682B4"],
                        [1.0, "aquamarine"],
                    ],
                }
            ],
        },
        "layout": {
            "margin": {"t": 50, "b": 50, "r": 50, "l": 50},
            "autosize": True,
            "font": {"size": 16},
            "paper_bgcolor": "rgba(0,0,0,0)",
        },
    }
    _backgroundColor = "rgba(0,0,0,0)"
    _gridColor = "rgb(176, 176, 176)"

    def __init__(
        self,
        xlabs: StrArrayLike,
        ylabs: StrArrayLike,
        titles: Optional[StrArrayLike] = None,
        zlabs: Optional[StrArrayLike] = None,
    ):
        """
        This function initializes the plot/subplots based on the inputs provided.

        Parameters
        ----------
        xlabs: StrArrayLike
            The x-axes labels for the plots
        ylabs: StrArrayLike
            The y-axes labels for the plots
        titles: StrArrayLike, optional
            The titles for the plots. (Default value = None)
        zlabs: StrArrayLike, optional
            The z-axes labels of for the plots. Setting this forces subplots to be 3D. (Default value = None)
        """

        # View distance used by the view method
        self.viewDistance = 2.5

        # Consistify all label types
        if isinstance(xlabs, np.ndarray):
            pass
        elif isinstance(xlabs, str):
            xlabs = np.array([[xlabs]])
        elif isinstance(xlabs, tuple) or isinstance(xlabs, list):
            xlabs = np.array(xlabs)
        else:
            TFCPrint.Error(
                "The xlabels provided are not of a valid type. Please provide valid xlabels"
            )
        if len(xlabs.shape) == 1:
            xlabs = np.expand_dims(xlabs, 1)

        if isinstance(ylabs, np.ndarray):
            pass
        elif isinstance(ylabs, str):
            ylabs = np.array([[ylabs]])
        elif isinstance(ylabs, tuple) or isinstance(ylabs, list):
            ylabs = np.array(ylabs)
        else:
            TFCPrint.Error(
                "The ylabels provided are not of a valid type. Please provide valid ylabels"
            )
        if len(ylabs.shape) == 1:
            ylabs = np.expand_dims(ylabs, 1)

        if zlabs is None:
            self._is3d = False
        else:
            self._is3d = True
            if isinstance(zlabs, np.ndarray):
                pass
            elif isinstance(zlabs, str):
                zlabs = np.array([[zlabs]])
            elif isinstance(zlabs, tuple) or isinstance(zlabs, list):
                zlabs = np.array(zlabs)
            else:
                TFCPrint.Error(
                    "The zlabels provided are not of a valid type. Please provide valid zlabels"
                )
            if len(zlabs.shape) == 1:
                zlabs = np.expand_dims(zlabs, 1)

        if titles is not None:
            if isinstance(titles, np.ndarray):
                titles = titles.flatten().tolist()
            elif isinstance(titles, str):
                titles = [titles]
            elif isinstance(titles, tuple) or isinstance(titles, list):
                titles = np.array(titles).flatten().tolist()
            else:
                TFCPrint.Error(
                    "The titles provided are not of a valid type. Please provide valid titles."
                )

        if int(np.prod(xlabs.shape)) != 1:
            from plotly.subplots import make_subplots

            if self._is3d:
                specs = [
                    [
                        {"is_3d": True},
                    ]
                    * xlabs.shape[1],
                ] * xlabs.shape[0]
                self.fig = make_subplots(
                    rows=xlabs.shape[0],
                    cols=xlabs.shape[1],
                    specs=specs,
                    subplot_titles=titles,
                )
            else:
                self.fig = make_subplots(
                    rows=xlabs.shape[0],
                    cols=xlabs.shape[1],
                    subplot_titles=titles,
                )
            self._hasSubplots = True
        else:
            self.fig = go.Figure()
            if titles is not None:
                self.fig["layout"]["title"] = titles[0]
            self._hasSubplots = False

        if self._hasSubplots:
            for row in range(xlabs.shape[0]):
                for col in range(xlabs.shape[1]):
                    if self._is3d:
                        if row == 0 and col == 0:
                            self.fig["layout"]["scene"]["xaxis"].update(
                                title=xlabs[row, col], exponentformat="e"
                            )
                            self.fig["layout"]["scene"]["yaxis"].update(
                                title=ylabs[row, col], exponentformat="e"
                            )
                            self.fig["layout"]["scene"]["zaxis"].update(
                                title=zlabs[row, col], exponentformat="e"
                            )
                        else:
                            self.fig["layout"]["scene" + str(row + col + 1)]["xaxis"].update(
                                title=xlabs[row, col], exponentformat="e"
                            )
                            self.fig["layout"]["scene" + str(row + col + 1)]["yaxis"].update(
                                title=ylabs[row, col], exponentformat="e"
                            )
                            self.fig["layout"]["scene" + str(row + col + 1)]["zaxis"].update(
                                title=zlabs[row, col], exponentformat="e"
                            )

                    else:
                        if row == 0 and col == 0:
                            self.fig["layout"]["xaxis"].update(
                                title=xlabs[row, col], exponentformat="e"
                            )
                            self.fig["layout"]["yaxis"].update(
                                title=ylabs[row, col], exponentformat="e"
                            )
                        else:
                            self.fig["layout"]["xaxis" + str(row + col + 1)].update(
                                title=xlabs[row, col], exponentformat="e"
                            )
                            self.fig["layout"]["yaxis" + str(row + col + 1)].update(
                                title=ylabs[row, col], exponentformat="e"
                            )

        else:
            if self._is3d:
                self.fig.update_layout(
                    scene=dict(xaxis=dict(title=xlabs[0, 0], exponentformat="e"))
                )
                self.fig.update_layout(
                    scene=dict(yaxis=dict(title=ylabs[0, 0], exponentformat="e"))
                )
                self.fig.update_layout(
                    scene=dict(zaxis=dict(title=zlabs[0, 0], exponentformat="e"))
                )
            else:
                self.fig.update_xaxes(title=xlabs[0, 0], exponentformat="e")
                self.fig.update_yaxes(title=ylabs[0, 0], exponentformat="e")

        # Update grid and background colors
        if self._is3d:
            for row in range(xlabs.shape[0]):
                for col in range(xlabs.shape[1]):
                    if row == 0 and col == 0:
                        self.fig["layout"]["scene"]["xaxis"]["gridcolor"] = self._gridColor
                        self.fig["layout"]["scene"]["yaxis"]["gridcolor"] = self._gridColor
                        self.fig["layout"]["scene"]["zaxis"]["gridcolor"] = self._gridColor
                        self.fig["layout"]["scene"]["xaxis"][
                            "backgroundcolor"
                        ] = self._backgroundColor
                        self.fig["layout"]["scene"]["yaxis"][
                            "backgroundcolor"
                        ] = self._backgroundColor
                        self.fig["layout"]["scene"]["zaxis"][
                            "backgroundcolor"
                        ] = self._backgroundColor
                    else:
                        self.fig["layout"]["scene" + str(row + col + 1)]["xaxis"].update(
                            gridcolor=self._gridColor,
                            zerolinecolor=self._gridColor,
                            backgroundcolor=self._backgroundColor,
                        )
                        self.fig["layout"]["scene" + str(row + col + 1)]["yaxis"].update(
                            gridcolor=self._gridColor,
                            zerolinecolor=self._gridColor,
                            backgroundcolor=self._backgroundColor,
                        )
                        self.fig["layout"]["scene" + str(row + col + 1)]["zaxis"].update(
                            gridcolor=self._gridColor,
                            zerolinecolor=self._gridColor,
                            backgroundcolor=self._backgroundColor,
                        )
        else:
            self.fig.update_xaxes(
                gridcolor=self._gridColor, linecolor="black", zerolinecolor=self._gridColor
            )
            self.fig.update_yaxes(
                gridcolor=self._gridColor, linecolor="black", zerolinecolor=self._gridColor
            )
            self.fig.update_layout(plot_bgcolor=self._backgroundColor)

        # Update layout
        self.fig.update_layout(
            template=self._template,
        )

    def Surface(
        self, row: Optional[uint] = None, col: Optional[uint] = None, **kwargs
    ) -> plotly.graph_objects.Surface:
        """
        Creates a plotly surface on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        row : Optional[uint]
            Subplot row (Default value = None)
        col : Optional[uint]
            Subplot column (Default value = None)
        **kwargs : dict, optional
            Keyword arguments passed on to plotly.graphic_objects.Surface

        Returns
        -------
        surf : plotly.graph_objects.Surface
        """
        return self.fig.add_trace(go.Surface(**kwargs), row=row, col=col)

    def Scatter3d(
        self, row: Optional[uint] = None, col: Optional[uint] = None, **kwargs
    ) -> plotly.graph_objects.Scatter3d:
        """
        Creates a 3d plotly scatter on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        row : Optional[uint]
             Subplot row (Default value = None)
        col : Optional[uint]
             Subplot column (Default value = None)
        **kwargs : dict, optional
            keyword arguments passed on to plotly.graphic_objects.Scatter3d

        Returns
        -------
        scatter : plotly.graph_objects.Scatter3d
        """
        return self.fig.add_trace(go.Scatter3d(**kwargs), row=row, col=col)

    def Scatter(
        self, row: Optional[uint] = None, col: Optional[uint] = None, **kwargs
    ) -> plotly.graph_objects.Scatter:
        """
        Creates a plotly scatter on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        row : Optional[uint]
             Subplot row (Default value = None)
        col : Optional[uint]
             Subplot column (Default value = None)
        **kwargs : dict, optional
            keyword arguments passed on to plotly.graph_objects.Scatter

        Returns
        -------
        scatter : plotly.graphic_objects.Scatter
        """
        return self.fig.add_trace(go.Scatter(**kwargs), row=row, col=col)

    def Histogram(
        self, row: Optional[uint] = None, col: Optional[uint] = None, **kwargs
    ) -> plotly.graph_objects.Histogram:
        """
        Creates a plotly histogram on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        row : Optional[uint]
             Subplot row (Default value = None)
        col : Optional[uint]
             Subplot column (Default value = None)
        **kwargs : dict, optional
            keyword arguments passed on to plotly.graph_objects.Histogram

        Returns
        -------
        hist : plotly.graphic_objects.Histogram
        """
        return self.fig.add_trace(go.Histogram(**kwargs), row=row, col=col)

    def Contour(
        self, row: Optional[uint] = None, col: Optional[uint] = None, **kwargs
    ) -> plotly.graph_objects.Contour:
        """
        Creates a plotly contour on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        row : Optional[uint]
             Subplot row (Default value = None)
        col : Optional[uint]
             Subplot column (Default value = None)
        **kwargs : dict, optional
            keyword arguments passed on to plotly.graphic_objects.Contour

        Returns
        -------
        contour : plotly.graph_objects.Contour
        """
        return self.fig.add_trace(go.Contour(**kwargs), row=row, col=col)

    def Box(
        self, row: Optional[uint] = None, col: Optional[uint] = None, **kwargs
    ) -> plotly.graph_objects.Box:
        """
        Creates a plotly box on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        row : Optional[uint]
             Subplot row (Default value = None)
        col : Optional[uint]
             Subplot column (Default value = None)
        **kwargs : dict, optional
            keyword arguments passed on to plotly.graph_objects.Box

        Returns
        -------
        box : plotly.graphic_objects.Box
        """
        return self.fig.add_trace(go.Box(**kwargs), row=row, col=col)

    def Violin(
        self, row: Optional[uint] = None, col: Optional[uint] = None, **kwargs
    ) -> plotly.graph_objects.Violin:
        """
        Creates a plotly violin on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        row : Optional[uint]
             Subplot row (Default value = None)
        col : Optional[uint]
             Subplot column (Default value = None)
        **kwargs : dict, optional
            keyword arguments passed on to plotly.graph_objects.Violin

        Returns
        -------
        violin : plotly.graphic_objects.Violin
        """
        return self.fig.add_trace(go.Violin(**kwargs), row=row, col=col)

    def Volume(
        self, row: Optional[uint] = None, col: Optional[uint] = None, **kwargs
    ) -> plotly.graph_objects.Volume:
        """
        Creates a plotly volume on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        row : Optional[uint]
             Subplot row (Default value = None)
        col : Optional[uint]
             Subplot column (Default value = None)
        **kwargs : dict, optional
            keyword arguments passed on to plotly.graph_objects.Volume

        Returns
        -------
        volume : plotly.graphic_objects.Volume
        """
        return self.fig.add_trace(go.Volume(**kwargs), row=row, col=col)

    def show(self, **kwargs):
        """
        Calls the figure's show method.

        Parameters
        ----------
        **kwargs : dict, optional
            keyword arguments passed on to fig.show
        """
        self.fig.show(**kwargs)

    def save(
        self,
        fileName: Path,
        tight: bool = True,
        fileType: Literal["pdf", "jpg", "png", "svg", "eps", "html", None] = None,
        **kwargs,
    ):
        """
        Saves the figure using the type specified. If HTML is specified, the figure will
        be saved as a dynamic html file. All other file types are static.

        Parameters
        ----------
        fileName : Path
            File name to save the figure as.
        tight : boolean, optional
            If the fileType is pdf or png and this value is true, then a tool is used to eliminate whitespace. pdfCropMargins is used for PDFs and PIL is used for png's. (Default value = True)
        fileType : Literal["pdf","jpg","png","svg","eps","html",None], optional
            File suffix to use. If None, then the suffix will be inferred from the suffix of the fileName. (Default value = None)
        **kwargs : dict, optional
            Keyword arguments passed onto fig.write_image or fig.write_html, depending on fileType.
        """
        if not fileType:
            from pathlib import Path

            suffix = Path(fileName).suffix[1:]
            if suffix in ["pdf", "jpg", "png", "svg", "eps", "html"]:
                fileType = suffix
            else:
                fileType = "png"
                TFCPrint.Warning(
                    f"Warning, file type could not be inferred from {fileName}. The file type has been set to png."
                )
                fileName += "." + fileType
            fileNameFull = fileName
        else:
            fileNameFull = fileName + "." + fileType
        if fileType == "html":
            self.fig.write_html(fileNameFull, **kwargs)
        elif fileType == "png" and tight == True:
            from io import BytesIO
            from PIL import Image

            dark = BytesIO()
            dark.write(self.fig.to_image(fileType, **kwargs))
            dark.seek(0)
            pilImage = Image.open(dark)
            pngArray = np.array(pilImage)
            blankPx = pngArray[0, 0, :]
            mask = pngArray != blankPx
            coords = np.argwhere(mask)
            x0, y0, z0 = coords.min(axis=0)
            x1, y1, z1 = coords.max(axis=0) + 1
            croppedBox = pngArray[x0:x1, y0:y1, z0:z1]
            pilImage = Image.fromarray(croppedBox, "RGBA")
            pilImage.save(fileNameFull)
        else:
            self.fig.write_image(fileNameFull, **kwargs)
            if fileType == "pdf" and tight:
                from pdfCropMargins import crop

                crop(["-p", "0", fileNameFull])

    def UploadToPlotly(self, username: str, apiKey: str, fileName: str, autoOpen: bool = False):
        """
        Upload your plot to Plotly.

        Parameters
        ----------
        username : str
            Plotly username
        apiKey : str
            Plotly api_key
        fileName : str
            Name of the file to save the plot as.
        autoOpen : bool, optional
            If true, plot will open in browser after saving. (Default value = False)
        """

        import chart_studio

        chart_studio.tools.set_credentials_file(username=username, api_key=apiKey)
        return chart_studio.plotly.plot(self.fig, filename=fileName, auto_open=autoOpen)

    def view(
        self,
        azimuth: float,
        elevation: float,
        row: Optional[uint] = None,
        col: Optional[uint] = None,
        viewDistance: Optional[float] = None,
    ):
        """
        Change the view on the subplot specified by row and col or on
        the main figure if not using subplots.

        Parameters
        ----------
        azimuth : float
            Azimuth value in degrees
        elevation : float
            Elevation value in degrees
        row : Optional[uint]
             Subplot row (Default value = None)
        col : Optional[uint]
             Subplot column (Default value = None)
        viewDistance : Optional[float]
            Distance from camera to plot. (Default value = self.viewDistance)
        """

        if not self._is3d:
            TFCPrint.Error("The view method is only for 3d plots.")

        if viewDistance:
            self.viewDistance = viewDistance

        azimuth *= np.pi / 180.0
        elevation *= np.pi / 180.0
        dark = self.viewDistance * np.array(
            [
                -np.sin(azimuth) * np.cos(elevation),
                -np.cos(azimuth) * np.cos(elevation),
                np.sin(elevation),
            ]
        )
        if row and col:
            sceneNum = row + col - 1
            if sceneNum == 1:
                sceneNum = ""
            else:
                sceneNum = str(sceneNum)

            self.fig["layout"]["scene" + sceneNum]["camera"].eye = dict(
                x=dark[0], y=dark[1], z=dark[2]
            )
        else:
            self.fig["layout"]["scene"]["camera"].eye = dict(x=dark[0], y=dark[1], z=dark[2])

    def FullScreen(self):
        """
        Make the plot full screen.
        """
        import tkinter as tk

        root = tk.Tk()
        root.withdraw()
        width = root.winfo_screenwidth()
        height = root.winfo_screenheight()

        self.fig.update_layout(width=width, height=height)

    def PartScreen(self, width: float, height: float, units: Literal["in", "mm", "px"] = "in"):
        """
        Make the plot size equal to width x height.

        Parameters
        ----------
        width : float
            Width of the plot
        height : float
            Height of the plot
        units : Literal["in","mm","px"], optional
            Units width and height are given in. (Default value = inches)
        """
        if units != "px":
            import tkinter as tk

            root = tk.Tk()
            root.withdraw()

            if units == "in":
                widthConvert = root.winfo_screenwidth() / (root.winfo_screenmmwidth() / 25.4)
                heightConvert = root.winfo_screenheight() / (root.winfo_screenmmheight() / 25.4)
            else:
                widthConvert = root.winfo_screenwidth() / root.winfo_screenmmwidth()
                heightConvert = root.winfo_screenheight() / root.winfo_screenmmheight()

            self.fig.update_layout(
                width=int(width * widthConvert), height=int(height * heightConvert)
            )
        else:
            self.fig.update_layout(width=width, height=height)

    def NormalizeColorScale(
        self,
        types: List[str] = [],
        data: Optional[str] = None,
        cmax: Optional[float] = None,
        cmin: Optional[float] = None,
    ):
        """
        Normalizes the color scale for the plots whose type is in the types list.

        If cmax and/or cmin are given, then the data variable is not used, and all
        plots whose type is in the types list are assigned that cmax/cmin value.

        If cmax and cmin are not specified, then they are set by taking the max and min
        of the data that matches the data variable in all plots whose type is in the
        typpes list.

        Parameters
        ----------
        types: List[str]
            Plot types to set cmax and cmin for.
        data: Optional[str]
            Data type to use to calculate cmax and cmin if not already specified. (Default value = None)
        cmax: Optional[float]
            cmax value to use when setting the colorscale
        cmin: Optional[float]
            cmin value to use when setting the colorscale
        """

        dark = self.fig.data
        if cmax is None and cmin is None:
            cmax = -np.inf
            cmin = np.inf
            for k in dark:
                if k.__class__.__name__ in types:
                    if data in k:
                        kmax = k[data].max()
                        kmin = k[data].min()

                    if kmax > cmax:
                        cmax = kmax
                    if kmin < cmin:
                        cmin = kmin

        colorScaleUpdate = {}
        if not cmax is None:
            colorScaleUpdate.update({"cmax": cmax})
        if not cmin is None:
            colorScaleUpdate.update({"cmin": cmin})

        for k in dark:
            if k.__class__.__name__ in types:
                if data in k:
                    k.update(colorScaleUpdate)



from jax._src.config import config

config.update("jax_enable_x64", True)

from copy import copy
import numpy as onp
import jax.numpy as np
from typing import cast
import numpy.typing as npt
from .utils.types import (
    Literal,
    uint,
    IntListOrArray,
    pint,
    NumberListOrArray,
    JaxOrNumpyArray,
    IntArrayLike,
    Array,
    Tuple,
)
from jax import core
from jax.interpreters import ad, batching, mlir
from jax.lib import xla_client
from jaxlib import hlo_helpers
import jaxlib.mlir.ir as ir

from .utils.TFCUtils import TFCPrint


class mtfc:
    """
    This is the multivariate TFC class. It acts as a container that holds:

    * The linear map from the domain of the DE to the domain of the free-function.
    * The necessary JAX code that enables automatic differentiation of the constrained experssion and Jacobians of the residual with respect to the unknown coefficients in the linear combination of basis functions that make up the free function.
    * Other useful TFC related functions such as collocation point creation.

    In addition, this class ties these methods together to form a utility that enables a higher level of code abstraction
    such that the end-user scripts are simple, clear, and elegant implementations of TFC.

    Parameters
    ----------
    n : IntListOrArray
        Number of points to use per-dimension when discretizing the domain. List or array must be same length as number of dimensions.

    nC : IntArrayLike
        Number of functions to remove from the beginning of free function linear expansion. This variable is used to account for basis functions that are linearly dependent on support functions used in the construction of the constrained expressions. The constraints for each dimension can be expressed in 1 of 2 ways. Note that a value of -1 is used to indicate no constraints exist for a particular dimension.

        1. As an integer. When expressed as an integer, the first nC basis functions are removed from the free function.
        2. As a list or array. When expressed as a list or array, the basis functions corresponding to the numbers given by the list or array are removed from the free function.

    deg : uint
        Degree of the basis function expansion.

    dim : pint
        Number of dimensions in the domain.

    basis : Literal["CP","LeP","FS","ELMTanh","ELMSigmoid","ELMSin","ELMSwish","ELMReLU"], optional
        This optional keyword argument specifies the basis functions to be used. (Default value = "CP")

    x0 : NumberListOrArray
        Specifies the beginning of the DE domain. (Default value = None)

    xf : NumberListOrArray
        Specifies the end of the DE domain. (Default value = None)

    backend : Literal["C++", "Python"]
        This optional keyword sets the backend used to compute the basis functions. The C++ can be used with JIT, but can only be used for doubles. The Python backend can be used for other field types, e.g., complex numbers, but does not have JIT translations. Instead, pejit must be used to set the basis function outputs as compile time constants in order to JIT.
    """

    def __init__(
        self,
        n: IntListOrArray,
        nC: IntArrayLike,
        deg: uint,
        dim: pint = 2,
        basis: Literal[
            "CP", "LeP", "FS", "ELMTanh", "ELMSigmoid", "ELMSin", "ELMSwish", "ELMReLU"
        ] = "CP",
        x0: NumberListOrArray = [],
        xf: NumberListOrArray = [],
        backend: Literal["C++", "Python"] = "C++",
    ):
        # Store givens
        self._elm_classes = ["ELMSigmoid", "ELMTanh", "ELMSin", "ELMSwish", "ELMReLU"]
        self.deg = deg
        self.dim = dim
        self._backend = backend

        _int_types = [onp.intp, onp.int8, onp.int16, onp.int32, onp.int64]

        # Set N based on user input
        if isinstance(n, np.ndarray):
            if not n.flatten().shape[0] == dim:
                TFCPrint.Error(
                    "n has length "
                    + str(n.flatten().shape[0])
                    + ", but it should be equal to the number of dimensions, "
                    + str(dim)
                    + "."
                )
            self.n = n.astype(np.int32)
        else:
            if not len(n) == dim:
                TFCPrint.Error(
                    "n has length "
                    + str(n)
                    + ", but it should be equal to the number of dimensions, "
                    + str(dim)
                    + "."
                )
            self.n = np.array(n, dtype=np.int32)
        self.N = int(np.prod(self.n, dtype=np.int32))

        self.basis = basis

        # Set x0 based on user input
        if x0 is None:
            self.x0 = onp.zeros(dim)
        else:
            if isinstance(x0, np.ndarray):
                if not x0.flatten().shape[0] == dim:
                    TFCPrint.Error(
                        "x0 has length "
                        + str(x0.flatten().shape[0])
                        + ", but it should be equal to the number of dimensions, "
                        + str(dim)
                        + "."
                    )
                self.x0 = cast(onp.ndarray, x0)
            else:
                if not len(x0) == dim:
                    TFCPrint.Error(
                        "x0 has length "
                        + str(len(x0))
                        + ", but it should be equal to the number of dimensions, "
                        + str(dim)
                        + "."
                    )
                self.x0 = onp.array(x0).flatten()
                if not self.x0.shape[0] == dim:
                    TFCPrint.Error(
                        "x0 has length "
                        + str(self.x0.shape[0])
                        + ", but it should be equal to the number of dimensions, "
                        + str(dim)
                        + "."
                    )

        if self.x0.dtype in _int_types:
            self.x0 = onp.array(self.x0, dtype=onp.float64)
            TFCPrint.Warning(
                "x0 is an integer type. Converting to float64 to avoid errors down the line."
            )

        # Set xf based on user input
        if xf is None:
            self.xf = onp.zeros(dim)
        else:
            if isinstance(xf, np.ndarray):
                if not xf.flatten().shape[0] == dim:
                    TFCPrint.Error(
                        "xf has length "
                        + str(xf.flatten().shape[0])
                        + ", but it should be equal to the number of dimensions, "
                        + str(dim)
                        + "."
                    )
                self.xf = cast(onp.ndarray, xf)
            else:
                if not len(xf) == dim:
                    TFCPrint.Error(
                        "xf has length "
                        + str(len(xf))
                        + ", but it should be equal to the number of dimensions, "
                        + str(dim)
                        + "."
                    )
                self.xf = onp.array(xf).flatten()
                if not self.xf.shape[0] == dim:
                    TFCPrint.Error(
                        "xf has length "
                        + str(self.xf.shape[0])
                        + ", but it should be equal to the number of dimensions, "
                        + str(dim)
                        + "."
                    )
        if self.xf.dtype in _int_types:
            self.xf = onp.array(self.xf, dtype=onp.float64)
            TFCPrint.Warning(
                "xf is an integer type. Converting to float64 to avoid errors down the line."
            )

        # Create nC matrix based on user input
        if basis in self._elm_classes:
            if isinstance(nC, int):
                self.nC = onp.arange(nC, dtype=onp.int32)
            elif isinstance(nC, np.ndarray):
                self.nC = cast(onp.ndarray, nC.astype(onp.int32))
            elif isinstance(nC, list):
                self.nC = onp.array(nC, dtype=np.int32)
            if self.nC.shape[0] > self.deg:
                TFCPrint.Error("Number of basis functions is less than number of constraints!")
            if np.any(self.nC < 0):
                TFCPrint.Error(
                    "To set nC to -1 (no constraints) either use nC = -1 or nC = 0 (i.e., use an integer not a list or array). Do not put only -1 in a list or array, this will cause issues in the C++ layer."
                )
        else:
            if isinstance(nC, int):
                TFCPrint.Error(
                    "Cannot use type int for nC when specifying non-ELM type basis function."
                )
            # Using explicit type casts here to keep LSPs happy. At this point, we know nC is not a regular integer.
            if isinstance(nC, np.ndarray) and len(nC.shape) > 1:
                if not nC.shape[0] == self.dim:
                    TFCPrint.Error(
                        "nC has "
                        + str(nC.flatten().shape[0])
                        + " rows, but the row number should be equal to the number of dimensions, "
                        + str(dim)
                        + "."
                    )
                self.nC = cast(onp.ndarray, nC.astype(np.int32))
            else:
                if isinstance(nC, np.ndarray):
                    nC = nC.tolist()
                if not len(cast(IntListOrArray, nC)) == dim:
                    TFCPrint.Error(
                        "nC has length "
                        + str(len(cast(IntListOrArray, nC)))
                        + ", but it should be equal to the number of dimensions, "
                        + str(dim)
                        + "."
                    )
                nCmax = 0
                for k in range(dim):
                    if isinstance(cast(IntListOrArray, nC)[k], np.ndarray):
                        nCk = np.array(cast(IntListOrArray, nC)[k]).flatten()
                    else:
                        nCk = np.array([cast(IntListOrArray, nC)[k]]).flatten()
                    if nCk.shape[0] == 1:
                        maxk = nCk[0]
                    else:
                        maxk = nCk.shape[0]
                    if maxk > nCmax:
                        nCmax = maxk

                if nCmax > self.deg:
                    TFCPrint.Error(
                        "Number of basis functions is less than the number of constraints!"
                    )

                onC = onp.zeros((dim, nCmax))
                for k in range(dim):
                    if isinstance(cast(IntListOrArray, nC)[k], np.ndarray):
                        nCk = np.array(cast(IntListOrArray, nC)[k]).flatten()
                    else:
                        nCk = onp.array([cast(IntListOrArray, nC)[k]]).flatten()
                    j = nCk.shape[0]
                    if j == 1:
                        nCk = onp.arange(nCk[0])
                        j = nCk.shape[0]
                    if j < nCmax:
                        if j == 0:
                            nCk = -1.0 * onp.ones(nCmax)
                        else:
                            nCk = cast(npt.NDArray, np.hstack([nCk, -1 * np.ones(nCmax - j)]))
                    onC[k, :] = nCk.astype(np.int32)
                self.nC = onp.array(onC.tolist(), dtype=np.int32)

        # Setup the basis function
        if backend == "C++":
            from .utils import BF
        elif backend == "Python":
            from .utils.BF import BF_Py as BF
        else:
            TFCPrint.Error(
                f'The backend {backend} was specified, but can only be one of "C++" or "Python".'
            )
        if self.basis == "CP":
            self.basisClass = BF.nCP(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = -1.0
            zf = 1.0
        elif self.basis == "LeP":
            self.basisClass = BF.nLeP(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = -1.0
            zf = 1.0
        elif self.basis == "FS":
            self.basisClass = BF.nFS(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = -np.pi
            zf = np.pi
        elif self.basis == "ELMSigmoid":
            self.basisClass = BF.nELMSigmoid(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        elif self.basis == "ELMTanh":
            self.basisClass = BF.nELMTanh(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        elif self.basis == "ELMSin":
            self.basisClass = BF.nELMSin(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        elif self.basis == "ELMSwish":
            self.basisClass = BF.nELMSwish(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        elif self.basis == "ELMReLU":
            self.basisClass = BF.nELMReLU(self.x0, self.xf, self.nC, self.deg + 1)
            z0 = 0.0
            zf = 1.0
        else:
            TFCPrint.Error("Invalid basis selection. Please select a valid basis")

        if self.basisClass.numBasisFunc > self.N:
            TFCPrint.Warning(
                "Warning, you have more basis functions than points!\nThis may lead to large solution errors!"
            )

        self.c = self.basisClass.c

        # Calculate z points and corresponding x
        self.z = onp.zeros((self.dim, self.N), dtype=self.x0.dtype)
        x = tuple([onp.zeros(self.N, dtype=self.x0.dtype) for _ in range(self.dim)])
        if self.basis in ["CP", "LeP"]:
            for k in range(self.dim):
                nProd = int(onp.prod(self.n[k + 1 :]))
                nStack = int(onp.prod(self.n[0:k]))
                j = self.n[k] - 1
                # Multiplying x0 by 0 here so the array will have the
                # same type as x0.
                I = onp.linspace(0 * x0[0], j, j + 1).reshape((j + 1, 1))
                dark = onp.cos(np.pi * (j - I) / float(j))
                dark = onp.hstack([dark] * nProd).flatten()
                self.z[k, :] = onp.array([dark] * nStack).flatten()
                x[k][:] = (self.z[k, :] - z0) / self.c[k] + self.x0[k]
        else:
            for k in range(self.dim):
                nProd = int(onp.prod(self.n[k + 1 :]))
                nStack = int(onp.prod(self.n[0:k]))
                dark = onp.linspace(z0, zf, num=self.n[k], dtype=self.x0.dtype).reshape(
                    (self.n[k], 1)
                )
                dark = onp.hstack([dark] * nProd).flatten()
                self.z[k, :] = onp.array([dark] * nStack).flatten()
                x[k][:] = (self.z[k, :] - z0) / self.c[k] + self.x0[k]

        self.z: Array = cast(Array, np.array(self.z.tolist()))
        self.x: Tuple[Array, ...] = tuple(
            [cast(Array, np.array(x[k].tolist())) for k in range(self.dim)]
        )

        self.SetupJAX()

    def H(self, *x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the basis function matrix for the points specified by *x.

        Parameters
        ----------
        *x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        H : NDArray
            Basis function matrix.
        """
        d = onp.zeros(self.dim, dtype=np.int32)
        return self._Hjax(*x, d=d, full=full)

    def Hx(self, *x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the derivative of the basis function matrix for the points specified by *x with respect to the first variable.

        Parameters
        ----------
        *x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        Hx : NDArray
            Derivative of the basis function matrix with respect to the first variable.
        """
        d = onp.zeros(self.dim, dtype=np.int32)
        d[0] = 1
        return self._Hjax(*x, d=d, full=full)

    def Hx2(self, *x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the second derivative of the basis function matrix for the points specified by *x with respect to the first variable.

        Parameters
        ----------
        *x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        Hx2 : NDArray
            Second derivative of the basis function matrix with respect to the first variable.
        """
        d = onp.zeros(self.dim, dtype=np.int32)
        d[0] = 2
        return self._Hjax(*x, d=d, full=full)

    def Hy2(self, *x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the second derivative of the basis function matrix for the points specified by *x with respect to the second variable.

        Parameters
        ----------
        *x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        Hy2 : NDArray
            Second derivative of the basis function matrix with respect to the second variable.
        """
        d = onp.zeros(self.dim, dtype=np.int32)
        d[1] = 2
        return self._Hjax(*x, d=d, full=full)

    def Hx2y(self, *x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the mixed derivative (second order derivative with respect to the first variable and first order with respect
        to the second variable) of the basis function matrix for the points specified by *x.

        Parameters
        ----------
        *x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        Hx2y : NDArray
            Mixed derivative of the basis function matrix with respect to the first variable.
        """
        d = onp.zeros(self.dim, dtype=np.int32)
        d[0] = 2
        d[1] = 1
        return self._Hjax(*x, d=d, full=full)

    def Hy(self, *x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the derivative of the basis function matrix for the points specified by *x with respect to the second variable.

        Parameters
        ----------
        *x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        Hy : NDArray
            Derivative of the basis function matrix with respect to the second variable.
        """
        d = onp.zeros(self.dim, dtype=np.int32)
        d[1] = 1
        return self._Hjax(*x, d=d, full=full)

    def Hxy(self, *x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the mixed derivative (first order derivative with respect to the first variable and first order with respect
        to the second variable) of the basis function matrix for the points specified by *x.

        Parameters
        ----------
        *x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        Hxy : NDArray
            Mixed derivative of the basis function matrix with respect to the first variable.
        """
        d = onp.zeros(self.dim, dtype=np.int32)
        d[0] = 1
        d[1] = 1
        return self._Hjax(*x, d=d, full=full)

    def Hz(self, *x: JaxOrNumpyArray, full: bool = False) -> npt.NDArray:
        """
        This function computes the derivative of the basis function matrix for the points specified by *x with respect to the third variable.

        Parameters
        ----------
        *x : JaxOrNumpyArray
            Points to calculate the basis functions at.

        full : bool, optional
            If true then the values specified by nC to the utfc class are ignored and all basis functions are computed. (Default value = False)

        Returns
        -------
        Hz : NDArray
            Derivative of the basis function matrix with respect to the third variable.
        """
        d = onp.zeros(self.dim, dtype=np.int32)
        d[2] = 1
        return self._Hjax(*x, d=d, full=full)

    def SetupJAX(self):
        """This function is used internally by TFC to setup autograd primatives and create desired behavior when taking derivatives of TFC constrained expressions."""

        # Helper variables
        d0 = onp.zeros(self.dim, dtype=np.int32)

        # Regiser XLA function
        if self._backend == "C++":
            obj = self.basisClass.xlaCapsule
            xlaName_str = "BasisFunc" + str(self.basisClass.identifier)
            xlaName = xlaName_str.encode("utf-8")
            xla_client.register_custom_call_target(xlaName, obj, platform="cpu")

        # Create Primitives
        H_p = core.Primitive("H")

        def Hjax(*x: JaxOrNumpyArray, d: npt.NDArray[onp.int32] = d0, full: bool = False):
            return cast(npt.NDArray, H_p.bind(*x, d=d, full=full))

        # Implicit translations
        def H_impl(*x: npt.NDArray, d: npt.NDArray[onp.int32] = d0, full: bool = False):
            return self.basisClass.H(np.array(x), d, full)

        H_p.def_impl(H_impl)

        # Define abstract evaluation
        def H_abstract_eval(
            *x, d: npt.NDArray[onp.int32] = d0, full: bool = False
        ) -> core.ShapedArray:
            if full:
                dim1 = self.basisClass.numBasisFuncFull
            else:
                dim1 = self.basisClass.numBasisFunc
            if len(x[0].shape) == 0:
                dims = (dim1,)
            else:
                dims = (x[0].shape[0], dim1)
            return core.ShapedArray(dims, x[0].dtype)

        H_p.def_abstract_eval(H_abstract_eval)

        if self._backend == "C++":
            # XLA compilation
            def default_layout(shape):
                return tuple(range(len(shape) - 1, -1, -1))

            def H_xla(ctx, *x, d: uint = 0, full: bool = False):
                x_type = ir.RankedTensorType(x[0].type)
                dims = x_type.shape
                dim0 = dims[0]
                if full:
                    dim1 = self.basisClass.numBasisFuncFull
                else:
                    dim1 = self.basisClass.numBasisFunc
                res_types, res_shapes = hlo_helpers.mk_result_types_and_shapes(
                    [((dim0, dim1), x_type.element_type)]
                )
                return hlo_helpers.custom_call(
                    xlaName_str,
                    result_types=res_types,
                    result_shapes=res_shapes,
                    operands=[
                        hlo_helpers.hlo_s32(self.basisClass.identifier),
                        hlo_helpers.hlo.ConcatenateOp(x, 0).result,
                        hlo_helpers.hlo_s32(d),
                        hlo_helpers.hlo_s32(self.dim),
                        mlir.ir_constant(full),
                        hlo_helpers.hlo_s32(dim0),
                        hlo_helpers.hlo_s32(dim1),
                    ],
                    operand_layouts=[
                        (),
                        default_layout((dim0 * len(x),)),
                        default_layout((len(d),)),
                        (),
                        (),
                        (),
                        (),
                    ],
                    result_layouts=[
                        default_layout((dim0, dim1)),
                    ],
                ).results

            mlir.register_lowering(H_p, H_xla, platform="cpu")

        # Batching translation
        def H_batch(vec, batch, d: npt.NDArray[onp.int32] = d0, full: bool = False):
            return Hjax(*vec, d=d, full=full), batch[0]

        batching.primitive_batchers[H_p] = H_batch

        # Jacobian vector translation
        def H_jvp(arg_vals, arg_tans, d: npt.NDArray[onp.int32] = d0, full: bool = False):
            n = len(arg_vals)
            flat = len(arg_vals[0].shape) == 1
            dim0 = arg_vals[0].shape[0]
            if full:
                dim1 = self.basisClass.numBasisFuncFull
            else:
                dim1 = self.basisClass.numBasisFunc
            out_tans = np.zeros((dim0, dim1))
            for k in range(n):
                if not (type(arg_tans[k]) is ad.Zero):
                    if type(arg_tans[k]) is batching.BatchTracer:
                        flag = onp.any(arg_tans[k].val != 0)
                    else:
                        flag = onp.any(arg_tans[k] != 0)
                    if flag:
                        dark = copy(d)
                        dark[k] += 1
                        if flat:
                            out_tans += Hjax(*arg_vals, d=dark, full=full) * np.expand_dims(
                                arg_tans[k], 1
                            )
                        else:
                            out_tans += Hjax(*arg_vals, d=dark, full=full) * arg_tans[k]
            return (Hjax(*arg_vals, d=d, full=full), out_tans)

        ad.primitive_jvps[H_p] = H_jvp

        self._Hjax = Hjax



from .TFCUtils import TFCPrint

TFCPrint()
from .TFCUtils import egradRobust as egrad
from .TFCUtils import egrad as egradSimple
from .TFCUtils import (
    TFCDict,
    TFCDictRobust,
    NLLS,
    ComponentConstraintGraph,
    NllsClass,
    step,
    LS,
    LsClass,
    pejit,
)
from .MakePlot import MakePlot
from . import Latex
from . import BF
from .CeSolver import CeSolver



import numpy as np
import numpy.typing as npt
import mayavi
from mayavi import mlab
from matplotlib import colors as mcolors
from .types import Dict, Tuple, Path, Ge, Le, Annotated, Literal
from typing import Optional, Any, Union, Generator, Callable
from .TFCUtils import TFCPrint

Color = Union[str, Tuple[float, float, float, float], npt.NDArray[np.float64]]
TFCPrint()


class MakePlot:
    """MakePlot class for Mayavi."""

    @staticmethod
    def _str_to_rgb(color: str) -> Tuple[float, float, float]:
        """Call matplotlib's colorConverter.to_rgb on input string.

        Parameters
        ----------
        color : str
            Color string

        Returns
        -------
        color_rgb : Tuple[float, float, float]
            3-tuple of the RGB for the color
        """
        return mcolors.colorConverter.to_rgb(color)

    @staticmethod
    def _str_to_rgba(color, alpha: Optional[float] = None) -> Tuple[float, float, float, float]:
        """Call matplotlib's colorConverter.to_rgba on input string.

        Parameters
        ----------
        color : str
            Color string

        alpha : float
            Alpha value to use in the return RGBA. If None, then the returned alpha = 1. (Default value = None)

        Returns
        -------
        color_rgba : Tuple[float, float, float, float]
            4-tuple of the RGB for the color
        """
        return mcolors.colorConverter.to_rgba(color, alpha=alpha)

    @staticmethod
    def _ProcessKwargs(**kwargs: Any) -> Dict[str, Any]:
        """This function effectively extends common mlab keywords.

        Parameters
        ----------
        **kwargs : Any
            keyword arguments

        Returns
        -------
        kwargs : Dict[str, any]
            Same as input keyword arguments but color has been transformed to an RGB if it was a string.
        """
        # Process color argument if it exists
        if "color" in kwargs:
            if isinstance(kwargs["color"], str):
                kwargs["color"] = MakePlot._str_to_rgb(kwargs["color"])

        return kwargs

    @staticmethod
    def ColormapGradient(c1: Color, c2: Color) -> npt.NDArray[np.uint8]:
        """Returns a Mayavi LUT for a gradient that linearly transforms from c1 to c2.

        Parameters
        ----------
        c1 : Color
            Input color as a string or 4-element RGBA list.

        c2 : Color
            Input color as a string or 4-element RGBA list.

        Returns
        -------
        lut : npt.NDArray[np.uint8]
            Mayavi LUT for a gradient that linearly transforms from c1 to c2.
        """
        if isinstance(c1, str):
            c1 = np.array(MakePlot._str_to_rgba(c1))
        if isinstance(c2, str):
            c2 = np.array(MakePlot._str_to_rgba(c2))
        if not isinstance(c1, np.ndarray):
            c1 = np.array(c1)
        if not isinstance(c2, np.ndarray):
            c2 = np.array(c2)
        c1 = c1 * 255
        c2 = c2 * 255
        return np.linspace(c1, c2, 256, dtype=np.uint8)

    @staticmethod
    def SetColormapGradient(obj: Any, c1: Color, c2: Color):
        """
        Applies a linear, gradient colormap to the object. The colors in the gradient are c1 and c2.

        Parameters
        ----------
        obj : Any
            mlab object to apply the colormap gradient to.
        c1 : Color
            Input color as a string or 4-element RGBA list.
        c2 : Color
            Input color as a string or 4-element RGBA list.
        """
        obj.module_manager.scalar_lut_manager.lut.table = MakePlot.ColormapGradient(c1, c2)

    def __init__(self):
        """
        This function initializes the plot.
        """

        self.fig = mlab.figure(bgcolor=(1.0, 1.0, 1.0))
        self.scene = self.fig.scene

    def FullScreen(self):
        """
        This function makes the plot fullscreen.
        """

        # Get screensize
        import tkinter as tk

        root = tk.Tk()
        width = root.winfo_screenwidth()
        height = root.winfo_screenheight()

        # Get dpi and set new figsize
        dpi = float(self.fig.get_dpi())
        mlab.figure(self.fig, size=(width / dpi, height / dpi))

    def PartScreen(self, width: float, height: float):
        """This function makes the plot width x height inches.

        Parameters
        ----------
        width : float
            Width of the plot in inches.

        height : float
            Height of the plot in inches.
        """
        mlab.figure(self.fig, size=(width, height))

    IntRange = Annotated[int, Ge(10), Le(100000)]

    def animate(
        self,
        animFunc: Callable[[], Generator[None, None, None]],
        outDir: Path = "MyMovie",
        fileName: str = "images",
        save: bool = True,
        delay: IntRange = 10,
    ) -> Optional[mayavi.tools.animator.Animator]:
        """

        Parameters
        ----------
        animFunc : Callable[[], Generator[None, None, None]]
            Function that modifies what is displayed on the plot
        outDir : Path, optional
            Directory to save frames in: only used if save = True. (Default value = "MyMovie")
        fileName : str, optional
            Filename for the frames: only used if save = True. (Default value = "images")
        save : bool, optional
            Controls whether the function saves the frames of the animation or not. (Default value = True)
        delay: IntRange in [10,100000], optional
            Controls the delay used by mlab.animate. (Default value = 10)

        Returns
        -------
        animator : Optional[mayavi.tools.animator.Animator]
            Returns an animator object only if save = False. Otherwise, there is no return value.
        """
        if save:
            import os

            if not os.path.exists(outDir):
                os.mkdir(outDir)
            k = 0

            iterable = animFunc()
            while next(iterable, -1) != -1:
                fileNameFull = "{}{:0>6d}".format(fileName, k)
                self.save(os.path.join(outDir, fileNameFull), fileType="png")
                k += 1

            print(
                "ffmpeg -r 60 -i ./{0}/{1}%06d.png -c:v libx264 -profile:v high -pix_fmt yuv420p ./{0}/MyMovie.mp4".format(
                    outDir, fileName
                )
            )
        else:
            return mlab.animate(func=animFunc, delay=delay)()

    def save(
        self,
        fileName: Path,
        fileType: Literal[
            "png",
            "jpg",
            "bmp",
            "tiff",
            "ps",
            "eps",
            "pdf",
            "rib",
            "oogl",
            "iv",
            "wrl",
            "vrml",
            "obj",
            "x3d",
            "pov",
            None,
        ] = None,
    ):
        """
        This function saves the figure.

        Parameters
        ----------
        fileName : Path
            Filename where the figure should be saved.
        fileType : Literal["png", "jpg", "bmp", "tiff", "ps", "eps", "pdf", "rib", "oogl", "iv", "wrl", "vrml", "obj", "x3d", "pov", None], optional
            File suffix to use. If None, then the suffix will be inferred from the file name. (Default value = None)
        """
        if not fileType:
            from pathlib import Path

            suffix = Path(fileName).suffix[1:]
            if suffix in [
                "png",
                "jpg",
                "bmp",
                "tiff",
                "ps",
                "eps",
                "pdf",
                "rib",
                "oogl",
                "iv",
                "wrl",
                "vrml",
                "obj",
                "x3d",
                "pov",
            ]:
                fileType = suffix
            else:
                fileType = "pdf"
                TFCPrint.Warning(
                    f"Warning, file type could not be inferred from {fileName}. The file type has been set to pdf."
                )
                fileName += "." + fileType
        else:
            fileName += "." + fileType
        mlab.savefig(fileName, figure=self.fig)

    def show(self):
        """Re-draw the class's figure."""
        return mlab.draw(figure=self.fig)

    @property
    def show_axes(self):
        """
        Adds axes indicator.
        """
        return self.scene.show_axes

    @show_axes.setter
    def show_axes(self, val: bool):
        """
        Set axes indicator.

        Parameters
        ----------
        val : bool
            True turns on the axes indicator and False turns off the axes indicator.
        """
        self.scene.show_axes = val

    def view(self, *args: Any, **kwargs: Any):
        """
        Call mlab's view on the class's figure.

        Parameters
        ----------
        *args : Any
            args passed on to view
        **kwargs : Any
            kwargs passed on to view
        """
        return mlab.view(*args, figure=self.fig, **kwargs)

    def points3d(self, *args: Any, **kwargs: Any) -> mayavi.modules.glyph.Glyph:
        """
        Call mlab's points3d on the class's figure.

        Parameters
        ----------
        *args : Any
            args passed on to points3d
        **kwargs : Any
            kwargs passed on to points3d after being processed by _ProcessKwargs.

        Returns:
        --------
        points3d : mayavi.modules.glyph.Glyph
            Glyphs corresponding to the points.
        """
        kwargs = MakePlot._ProcessKwargs(**kwargs)
        return mlab.points3d(*args, figure=self.fig, **kwargs)

    def plot3d(self, *args: Any, **kwargs: Any) -> mayavi.modules.surface.Surface:
        """Call mlab's plot3d on the class's figure.

        Parameters
        ----------
        *args : Any
            args passed on to plot3d
        **kwargs : Any
            kwargs passed on to plot3d after being processed by _ProcessKwargs.

        Returns:
        --------
        surf : mayavi.modules.surface.Surface
            Line as a Mayavi surface.
        """
        kwargs = MakePlot._ProcessKwargs(**kwargs)
        return mlab.plot3d(*args, figure=self.fig, **kwargs)

    def surf(self, *args: Any, **kwargs: Any) -> mayavi.modules.surface.Surface:
        """Call mlab's surf on the class's figure.

        Parameters
        ----------
        *args : Any
            args passed on to surf
        **kwargs : Any
            kwargs passed on to surf after being processed by _ProcessKwargs.

        Returns:
        --------
        surf : mayavi.modules.surface.Surface
            Surface.
        """
        kwargs = MakePlot._ProcessKwargs(**kwargs)
        return mlab.surf(*args, figure=self.fig, **kwargs)

    def quiver3d(self, *args: Any, **kwargs: Any) -> mayavi.modules.vectors.Vectors:
        """
        Call mlab's quiver3d on the class's figure.

        Parameters
        ----------
        *args : Any
            args passed on to quiver3d
        **kwargs : Any
            kwargs passed on to quiver3d after being processed by _ProcessKwargs.

        Returns:
        --------
        vectors : mayavi.modules.vectors.Vectors
            Vectors associated with the quiver3d plot.
        """
        kwargs = MakePlot._ProcessKwargs(**kwargs)
        return mlab.quiver3d(*args, figure=self.fig, **kwargs)

    def mesh(self, *args: Any, **kwargs: Any) -> mayavi.modules.surface.Surface:
        """Call mlab's mesh on the class's figure.

        Parameters
        ----------
        *args : Any
            args passed on to mesh
        **kwargs : Any
            kwargs passed on to mesh after being processed by _ProcessKwargs.

        Returns:
        --------
        mesh : mayavi.modules.surface.Surface
            Mesh as a mayavi surface.
        """
        kwargs = MakePlot._ProcessKwargs(**kwargs)
        return mlab.mesh(*args, figure=self.fig, **kwargs)

    def contour3d(self, *args: Any, **kwargs: Any) -> mayavi.modules.iso_surface.IsoSurface:
        """
        Call mlab's contour3d on the class's figure.

        Parameters
        ----------
        *args : Any
            args passed on to contour3d
        **kwargs : Any
            kwargs passed on to mesh after being processed by _ProcessKwargs.

        Returns:
        --------
        contour : mayavi.modules.iso_surface.IsoSurface
            Coutour.
        """
        kwargs = MakePlot._ProcessKwargs(**kwargs)
        return mlab.contour3d(*args, figure=self.fig, **kwargs)



import sys
from typing import Union, Any, Callable
import numpy as np
import numpy.typing as npt
from jax import Array
from sympy.core.function import AppliedUndef
from sympy import Expr

if sys.version_info >= (3, 8):
    from typing import Literal, Protocol, TypedDict
else:
    from typing_extensions import Literal, Protocol, TypedDict

if sys.version_info >= (3, 9):
    from typing import Annotated

    List = list
    Tuple = tuple
    Dict = dict
    Tuple = tuple
else:
    from typing_extensions import Annotated
    from typing import List, Tuple, Dict, Tuple

from annotated_types import Gt, Ge, Lt, Le

# Path
# Path = Union[str, os.PathLike]
Path = str

# Integer > 0
pint = Annotated[int, Gt(0)]

# Integer >= 0
uint = Annotated[int, Ge(0)]

# General number type
Number = Union[int, float, complex]

if sys.version_info >= (3, 8):
    from numpy._typing._array_like import _ArrayLikeStr_co, _ArrayLikeInt_co

    # Array-like of strings
    StrArrayLike = _ArrayLikeStr_co

    # Array-like of integers
    IntArrayLike = _ArrayLikeInt_co
else:
    # Hacks to keep things working for Python 3.7
    StrArrayLike = Any
    IntArrayLike = Any

# List or array like
NumberListOrArray = Union[Tuple[Number, ...], List[Number], npt.NDArray[Any], Array]

# List or array of integers
IntListOrArray = Union[
    Tuple[int, ...],
    List[int],
    npt.NDArray[np.int32],
    npt.NDArray[np.int64],
    npt.NDArray[np.int16],
    npt.NDArray[np.int8],
]

# JAX array or numpy array
JaxOrNumpyArray = Union[npt.NDArray, Array]

# Tuple or list of array
TupleOrListOfArray = Union[Tuple[JaxOrNumpyArray, ...], List[JaxOrNumpyArray]]
TupleOrListOfNumpyArray = Union[Tuple[npt.NDArray, ...], List[npt.NDArray]]

# Sympy constraint operator
# Adding in Any here since sympy types are a bit funky at the moment
ConstraintOperator = Callable[[Union[AppliedUndef, Expr, Any]], Union[AppliedUndef, Any]]
ConstraintOperators = Union[List[ConstraintOperator], Tuple[ConstraintOperator, ...]]

# List or tuple of sympy expressions
# Adding in Any here since sympy types are a bit funky at the moment
Exprs = Union[List[Union[Expr, Any]], Tuple[Union[Expr, Any], ...]]



import sys

if sys.version_info >= (3, 9):
    Tuple = tuple
else:
    from typing import Tuple

__version__ = "1.0.10"


def _version_as_tuple(version_str: str) -> Tuple[int, ...]:
    return tuple(int(i) for i in version_str.split(".") if i.isdigit())


__version_info__ = _version_as_tuple(__version__)



import os
import pickle
import numpy as np

import matplotlib.pyplot as plt

from .TFCUtils import TFCPrint
from .types import StrArrayLike, Path, List, Dict, Literal, pint
from typing import Optional, Union, Generator, Callable

TFCPrint()


class MakePlot:
    """
    This class is used to easily create journal-article-ready plots and subplots. The class can create 2D as well as 3D plots
    and even has support for twin y-axes.
    """

    def __init__(
        self,
        xlabs: StrArrayLike,
        ylabs: StrArrayLike,
        titles: Optional[StrArrayLike] = None,
        twinYlabs: Optional[StrArrayLike] = None,
        zlabs: Optional[StrArrayLike] = None,
        style: Optional[Union[str, Dict, Path, List[str], List[Dict], List[Path]]] = None,
    ):
        """
        This function initializes the plot/subplots based on the inputs provided.

        Parameters
        ----------
        xlabs: StrArrayLike
            The x-axes labels for the plots
        ylabs: StrArrayLike
            The y-axes labels for the plots
        titles: StrArrayLike, optional
            The titles for the plots. (Default value = None)
        twinYlabs: StrArrayLike, optional
            The twin y-axes labels for the plots. Setting this forces twin axis y-axes. (Default value = None)
        zlabs: StrArrayLike, optional
            The z-axes labels of for the plots. Setting this forces subplots to be 3D. (Default value = None)
        style : Union[str, Dict, Path, List[str], List[Dict], List[Path]]
            Matplotlib style. (Default value = None)
        """

        # Apply a style if specified
        if style:
            plt.style.use(style)

        # Set the fontsizes and family
        smallSize = 16
        mediumSize = 18
        largeSize = 18
        plt.rc("font", size=smallSize)
        plt.rc("axes", titlesize=mediumSize)
        plt.rc("axes", labelsize=largeSize)
        plt.rc("xtick", labelsize=mediumSize)
        plt.rc("ytick", labelsize=mediumSize)
        plt.rc("legend", fontsize=smallSize)
        plt.rc("figure", titlesize=largeSize)

        # Create figure and store basic labels
        self.fig = plt.figure()

        # Consistify all label types
        if isinstance(xlabs, np.ndarray):
            pass
        elif isinstance(xlabs, str):
            xlabs = np.array([[xlabs]])
        elif isinstance(xlabs, tuple) or isinstance(xlabs, list):
            xlabs = np.array(xlabs)
        else:
            TFCPrint.Error(
                "The xlabels provided are not of a valid type. Please provide valid xlabels"
            )
        if len(xlabs.shape) == 1:
            xlabs = np.expand_dims(xlabs, 1)

        if isinstance(ylabs, np.ndarray):
            pass
        elif isinstance(ylabs, str):
            ylabs = np.array([[ylabs]])
        elif isinstance(ylabs, tuple) or isinstance(ylabs, list):
            ylabs = np.array(ylabs)
        else:
            TFCPrint.Error(
                "The ylabels provided are not of a valid type. Please provide valid ylabels"
            )
        if len(ylabs.shape) == 1:
            ylabs = np.expand_dims(ylabs, 1)

        if not zlabs is None:
            if isinstance(zlabs, np.ndarray):
                pass
            elif isinstance(zlabs, str):
                zlabs = np.array([[zlabs]])
            elif isinstance(zlabs, tuple) or isinstance(zlabs, list):
                zlabs = np.array(zlabs)
            else:
                TFCPrint.Error(
                    "The zlabels provided are not of a valid type. Please provide valid zlabels"
                )
            if len(zlabs.shape) == 1:
                zlabs = np.expand_dims(zlabs, 1)

        if titles is not None:
            if isinstance(titles, np.ndarray):
                pass
            elif isinstance(titles, str):
                titles = np.array([[titles]])
            elif isinstance(titles, tuple) or isinstance(titles, list):
                titles = np.array(titles)
            else:
                TFCPrint.Error(
                    "The titles provided are not of a valid type. Please provide valid titles."
                )
            if len(titles.shape) == 1:
                titles = np.expand_dims(titles, 1)

        if twinYlabs is not None:
            if isinstance(twinYlabs, np.ndarray):
                pass
            elif isinstance(twinYlabs, str):
                twinYlabs = np.array([[twinYlabs]])
            elif isinstance(twinYlabs, tuple) or isinstance(twinYlabs, list):
                twinYlabs = np.array(twinYlabs)
            else:
                TFCPrint.Error(
                    "The twin ylabels provided are not of a valid type. Please provide valid twin ylabels"
                )
            if len(twinYlabs.shape) == 1:
                twinYlabs = np.expand_dims(twinYlabs, 1)

        # Create all subplots and add labels
        if zlabs is None:
            n = xlabs.shape
            self.ax = list()
            count = 0
            for j in range(n[0]):
                for k in range(n[1]):
                    if xlabs[j, k] is None:
                        continue
                    self.ax.append(self.fig.add_subplot(n[0], n[1], j * n[1] + k + 1))
                    self.ax[count].set_xlabel(xlabs[j, k])
                    self.ax[count].set_ylabel(ylabs[j, k])
                    count += 1
        else:
            n = xlabs.shape
            self.ax = list()
            count = 0
            for j in range(n[0]):
                for k in range(n[1]):
                    if xlabs[j, k] is None:
                        continue
                    self.ax.append(
                        self.fig.add_subplot(n[0], n[1], j * n[1] + k + 1, projection="3d")
                    )
                    self.ax[count].set_xlabel(xlabs[j, k])
                    self.ax[count].set_ylabel(ylabs[j, k])
                    self.ax[count].set_zlabel(zlabs[j, k])
                    count += 1

        if twinYlabs is not None:
            self.twinAx = list()
            count = 0
            for j in range(n[0]):
                for k in range(n[1]):
                    if xlabs[j, k] is None:
                        continue
                    self.twinAx.append(self.ax[count].twinx())
                    self.twinAx[count].set_ylabel(twinYlabs[j, k])
                    count += 1

        # Add titles if desired
        if titles is not None:
            count = 0
            for j in range(n[0]):
                for k in range(n[1]):
                    if titles[j, k] is None:
                        continue
                    self.ax[count].set_title(titles[j, k])
                    count += 1

        # Set tight layout for the figure
        self.fig.tight_layout()

    def FullScreen(self):
        """
        This function makes the plot fullscreen.
        """

        # Get screensize
        import tkinter as tk

        root = tk.Tk()
        root.withdraw()
        width = root.winfo_screenwidth()
        height = root.winfo_screenheight()

        # Get dpi and set new figsize
        dpi = float(self.fig.get_dpi())
        self.fig.set_size_inches(width / dpi, height / dpi)

    def PartScreen(self, width: float, height: float):
        """
        This function makes the plot width x height inches.

        Parameters
        ----------
        width : float
            Width of the plot in inches.

        height : float
            Height of the plot in inches.
        """
        self.fig.set_size_inches(width, height)

    def show(self):
        """
        This function shows the plot.
        """
        self.fig.show()

    def draw(self):
        """
        This function draws the canvas.
        """
        self.fig.canvas.draw()

    def save(
        self,
        fileName: Path,
        transparent: bool = True,
        fileType: Literal["png", "pdf", "ps", "eps", "svg", None] = None,
    ):
        """
        This function crops and saves the figure.

        Parameters
        ----------
        fileName : Path
            Filename where the figure should be saved.
        transparent : bool, optional
            Whether to save the plot with transparency or not. (Default value = True)
        fileType : Literal["png", "pdf", "ps", "eps", "svg", None], optional
            File suffix to use. If None, then the suffix will inferred from the suffix of the fileName. (Default value = None)
        """
        if not fileType:
            from pathlib import Path

            suffix = Path(fileName).suffix[1:]
            if suffix in ["png", "pdf", "ps", "eps", "svg"]:
                fileType = suffix
            else:
                fileType = "pdf"
                TFCPrint.Warning(
                    f"Warning, file type could not be inferred from {fileName}. The file type has been set to pdf."
                )
        else:
            fileName += "." + fileType

        self.fig.savefig(
            fileName,
            bbox_inches="tight",
            pad_inches=0.05,
            dpi=300,
            format=fileType,
            transparent=transparent,
        )

    def savePickle(self, fileName: Path):
        """This function saves the figure in a pickle format so it can be opened and modified later.

        Parameters
        ----------
        fileName : Path
            Filename where the figure should be saved. Note, this should not include the file suffix.
        """
        pickle.dump(self, open(fileName + ".pickle", "wb"))

    def saveAll(
        self,
        fileName: Path,
        transparent: bool = True,
        fileType: Literal["png", "pdf", "ps", "eps", "svg", None] = None,
    ):
        """This function invokes the save and savePickle functions.

        Parameters
        ----------
        fileName : Path
            Filename where the figure should be saved.
        transparent : bool, optional
            Whether to save the plot with transparency or not. (Default value = True)
        fileType : Literal["png", "pdf", "ps", "eps", "svg", None], optional
            File suffix to use. If None, then the suffix will be inferred from the suffix of the fileName. (Default value = None)
        """
        self.save(fileName, transparent=transparent, fileType=fileType)
        self.savePickle(fileName)

    def animate(
        self,
        animFunc: Callable[[], Generator[None, None, None]],
        outDir: Path = "MyMovie",
        fileName: str = "images",
        save: bool = True,
        delay: pint = 10,
    ):
        """
        Creates an animation using a Python generator.

        Parameters
        ----------
        animFunc: Callable[[], Generator[None, None, None]]
            Function whose returned generator modifies what is displayed on the plot
        outDir : Path, optional
             Directory to save frames in: only used if save = True. (Default value = "MyMovie")
        fileName : str, optional
             Filename for the frames: only used if save = True. (Default value = "images")
        save : bool, optional
             Controls whether the function saves the frames of the animation or not. (Default value = True)
        delay : pint, optional
             Delay in milliseconds between frames: only used if save = False. (Default value = 10)
        """

        iterable = animFunc()

        if save:
            if not os.path.exists(outDir):
                os.mkdir(outDir)
            k = 0

            while next(iterable, -1) != -1:
                plt.pause(delay / 1000.0)
                fileNameFull = "{}{:0>6d}".format(fileName, k)
                self.save(os.path.join(outDir, fileNameFull), fileType="png")
                k += 1

            print(
                'ffmpeg -r 60 -i ./{0}/{1}%06d.png -c:v libx264 -profile:v high -pix_fmt yuv420p -vf "scale=trunc(iw/2)*2:trunc(ih/2)*2" ./{0}/MyMovie.mp4'.format(
                    outDir, fileName
                )
            )
        else:
            while next(iterable, -1) != -1:
                plt.pause(delay / 1000.0)



""" This is a dummy file used only to avoid errors in ReadTheDocs. The real BF.py is created during the setup once swig is run. """


def CP():
    pass


def LeP():
    pass


def LaP():
    pass


def HoPpro():
    pass


def HoPphy():
    pass


def FS():
    pass


def ELMReLU():
    pass


def ELMSigmoid():
    pass


def ELMTanh():
    pass


def ELMSin():
    pass


def ELMSwish():
    pass


def nCP():
    pass


def nLeP():
    pass


def nFS():
    pass


def nELMReLU():
    pass


def nELMSigmoid():
    pass


def nELMTanh():
    pass


def nELMSin():
    pass


def nELMSwish():
    pass



import numpy as np
import jax.numpy as jnp
from abc import ABC, abstractmethod
from numpy import typing as npt
from tfc.utils.types import uint, Number
from typing import Callable, Tuple


class BasisFunc(ABC):
    """
    Python implementation of the basis function classes. These are an alternative
    to the C++ versions. They can not be JIT-ed, but they do support alternative
    types, e.g., single float, complex, etc. Even though they cannnot be JIT-ed,
    they can often be used in JIT functions, if their arguments can be removed
    from said functions. For example, when solving an ODE, oftentimes the basis
    functions can be treated as compile time constants. This can be done using
    `pejit`: see `pejit` for more details.
    """

    def __init__(
        self,
        x0: Number,
        xf: Number,
        nC: npt.NDArray,
        m: uint,
        z0: Number = 0,
        zf: Number = float("inf"),
    ) -> None:
        """
        Initialize the basis class.

        Parameters
        ----------
        x0 : Number
            Start of the problem domain.
        xf : Number
            End of the problem domain.
        nC : npt.NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        z0 : Number
            Start of the basis function domain.
        zf : Number
            End of the basis function domain.
        """

        self._m = m
        self._nC = nC
        self._numC = len(nC)

        self._z0 = z0
        if zf == float("inf"):
            self._c = 1.0
            self._x0 = 0.0
        else:
            self._x0 = x0
            self._c = (zf - z0) / (xf - x0)

    def H(self, x: npt.NDArray, d: uint = 0, full: bool = False) -> npt.NDArray:
        """
        Returns the basis function matrix for the x with a derivative of order d.

        Parameters
        ----------
        x : NDArray
            Input array. Values to calculate the basis function for.
        d : uint
            Order of the derivative
        full : bool
            Whether to return the full basis function set, or remove
            the columns associated with self._nC.

        Returns
        -------
        H : NDArray
            The basis function values.
        """

        z = (x - self._x0) * self._c + self._z0
        if len(z.shape) == 1:
            z = np.expand_dims(z, 1)
        dMult = self._c**d
        F = self._Hint(z, d) * dMult
        if not full and self._numC > 0:
            F = np.delete(F, self._nC, axis=1)
        return F

    @abstractmethod
    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the basis function value.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """
        pass

    @property
    def c(self) -> Number:
        """
        Return the constants that map the problem domain to the basis
        function domain.

        Returns
        -------
        float
            The constant that maps the problem domain to the basis function
            domain.
        """

        return self._c


class CP(BasisFunc):
    """
    Chebyshev polynomial basis functions.
    """

    def __init__(
        self,
        x0: Number,
        xf: Number,
        nC: npt.NDArray,
        m: uint,
    ) -> None:
        """
        Initialize the basis class.

        Parameters
        ----------
        x0 : Number
            Start of the problem domain.
        xf : Number
            End of the problem domain.
        nC : npt.NDArray
            Basis functions to be removed
        m:  uint
            Number of basis functions.
        """
        super().__init__(x0, xf, nC, m, -1.0, 1.0)

    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the CP basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """
        N = np.size(z)
        One = np.ones_like(z)
        Zero = np.zeros_like(z)
        if self._m == 1:
            if d > 0:
                F = Zero
            else:
                F = One
            return F
        elif self._m == 2:
            if d > 1:
                F = np.hstack((Zero, Zero))
            elif d > 0:
                F = np.hstack((Zero, One))
            else:
                F = np.hstack((One, z))
            return F
        else:
            F = np.hstack((One, z, np.zeros((N, self._m - 2), dtype=z.dtype)))
            for k in range(2, self._m):
                F[:, k : k + 1] = 2 * z * F[:, k - 1 : k] - F[:, k - 2 : k - 1]

            def Recurse(dark: npt.NDArray, d: uint, dCurr: uint = 0) -> npt.NDArray:
                """
                Take derivative recursively.
                """
                if dCurr == d:
                    return dark
                else:
                    if dCurr == 0:
                        dark2 = np.hstack((Zero, One, np.zeros((N, self._m - 2), dtype=z.dtype)))
                    else:
                        dark2 = np.zeros((N, self._m), dtype=z.dtype)
                    for k in range(2, self._m):
                        dark2[:, k : k + 1] = (
                            (2 + 2 * dCurr) * dark[:, k - 1 : k]
                            + 2 * z * dark2[:, k - 1 : k]
                            - dark2[:, k - 2 : k - 1]
                        )
                    dCurr += 1
                    return Recurse(dark2, d, dCurr=dCurr)

            return Recurse(F, d)


class LeP(BasisFunc):
    """
    Legendre polynomial basis functions.
    """

    def __init__(
        self,
        x0: Number,
        xf: Number,
        nC: npt.NDArray,
        m: uint,
    ) -> None:
        """
        Initialize the basis class.

        Parameters
        ----------
        x0 : Number
            Start of the problem domain.
        xf : Number
            End of the problem domain.
        nC : npt.NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        """
        super().__init__(x0, xf, nC, m, -1.0, 1.0)

    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the LeP basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """
        N = np.size(z)
        One = np.ones_like(z)
        Zero = np.zeros_like(z)
        if self._m == 1:
            if d > 0:
                F = Zero
            else:
                F = One
            return F
        elif self._m == 2:
            if d > 1:
                F = np.hstack((Zero, Zero))
            elif d > 0:
                F = np.hstack((Zero, One))
            else:
                F = np.hstack((One, z))
            return F
        else:
            F = np.hstack((One, z, np.zeros((N, self._m - 2), dtype=z.dtype)))
            for k in range(1, self._m - 1):
                F[:, k + 1 : k + 2] = (
                    (2.0 * k + 1.0) * z * F[:, k : k + 1] - k * F[:, k - 1 : k]
                ) / (k + 1.0)

            def Recurse(dark: npt.NDArray, d: uint, dCurr: uint = 0) -> npt.NDArray:
                """
                Take derivative recursively.
                """
                if dCurr == d:
                    return dark
                else:
                    if dCurr == 0:
                        dark2 = np.hstack((Zero, One, np.zeros((N, self._m - 2), dtype=z.dtype)))
                    else:
                        dark2 = np.zeros((N, self._m), dtype=z.dtype)
                    for k in range(1, self._m - 1):
                        dark2[:, k + 1 : k + 2] = (
                            (2.0 * k + 1.0)
                            * ((dCurr + 1.0) * dark[:, k : k + 1] + z * dark2[:, k : k + 1])
                            - k * dark2[:, k - 1 : k]
                        ) / (k + 1.0)
                    dCurr += 1
                    return Recurse(dark2, d, dCurr=dCurr)

            return Recurse(F, d)


class LaP(BasisFunc):
    """
    Laguerre polynomial basis functions.
    """

    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the LaP basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """
        N = np.size(z)
        One = np.ones_like(z)
        Zero = np.zeros_like(z)
        if self._m == 1:
            if d > 0:
                F = Zero
            else:
                F = One
            return F
        elif self._m == 2:
            if d > 1:
                F = np.hstack((Zero, Zero))
            elif d > 0:
                F = np.hstack((Zero, -One))
            else:
                F = np.hstack((One, 1.0 - z))
            return F
        else:
            F = np.hstack((One, 1.0 - z, np.zeros((N, self._m - 2), dtype=z.dtype)))
            for k in range(1, self._m - 1):
                F[:, k + 1 : k + 2] = (
                    (2.0 * k + 1.0 - z) * F[:, k : k + 1] - k * F[:, k - 1 : k]
                ) / (k + 1.0)

            def Recurse(dark: npt.NDArray, d: uint, dCurr: uint = 0) -> npt.NDArray:
                """
                Take derivative recursively.
                """
                if dCurr == d:
                    return dark
                else:
                    if dCurr == 0:
                        dark2 = np.hstack((Zero, -One, np.zeros((N, self._m - 2), dtype=z.dtype)))
                    else:
                        dark2 = np.zeros((N, self._m), dtype=z.dtype)
                    for k in range(1, self._m - 1):
                        dark2[:, k + 1 : k + 2] = (
                            (2.0 * k + 1.0 - z) * dark2[:, k : k + 1]
                            - (dCurr + 1.0) * dark[:, k : k + 1]
                            - k * dark2[:, k - 1 : k]
                        ) / (k + 1.0)
                    dCurr += 1
                    return Recurse(dark2, d, dCurr=dCurr)

            return Recurse(F, d)


class HoPpro(BasisFunc):
    """
    Hermite probablist polynomial basis functions.
    """

    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the HoPpro basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function valuesa
        """
        N = np.size(z)
        One = np.ones_like(z)
        Zero = np.zeros_like(z)
        if self._m == 1:
            if d > 0:
                F = Zero
            else:
                F = One
            return F
        elif self._m == 2:
            if d > 1:
                F = np.hstack((Zero, Zero))
            elif d > 0:
                F = np.hstack((Zero, One))
            else:
                F = np.hstack((One, z))
            return F
        else:
            F = np.hstack((One, z, np.zeros((N, self._m - 2), dtype=z.dtype)))
            for k in range(1, self._m - 1):
                F[:, k + 1 : k + 2] = z * F[:, k : k + 1] - k * F[:, k - 1 : k]

            def Recurse(dark: npt.NDArray, d: uint, dCurr: uint = 0) -> npt.NDArray:
                """
                Take derivative recursively.
                """
                if dCurr == d:
                    return dark
                else:
                    if dCurr == 0:
                        dark2 = np.hstack((Zero, One, np.zeros((N, self._m - 2), dtype=z.dtype)))
                    else:
                        dark2 = np.zeros((N, self._m), dtype=z.dtype)
                    for k in range(1, self._m - 1):
                        dark2[:, k + 1 : k + 2] = (
                            (dCurr + 1.0) * dark[:, k : k + 1]
                            + z * dark2[:, k : k + 1]
                            - k * dark2[:, k - 1 : k]
                        )
                    dCurr += 1
                    return Recurse(dark2, d, dCurr=dCurr)

            return Recurse(F, d)


class HoPphy(BasisFunc):
    """
    Hermite physicist polynomial basis functions.
    """

    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the HoPpro basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function valuesa
        """
        N = np.size(z)
        One = np.ones_like(z)
        Zero = np.zeros_like(z)
        if self._m == 1:
            if d > 0:
                F = Zero
            else:
                F = One
            return F
        elif self._m == 2:
            if d > 1:
                F = np.hstack((Zero, Zero))
            elif d > 0:
                F = np.hstack((Zero, 2.0 * One))
            else:
                F = np.hstack((One, 2.0 * z))
            return F
        else:
            F = np.hstack((One, 2.0 * z, np.zeros((N, self._m - 2), dtype=z.dtype)))
            for k in range(1, self._m - 1):
                F[:, k + 1 : k + 2] = 2.0 * z * F[:, k : k + 1] - 2.0 * k * F[:, k - 1 : k]

            def Recurse(dark: npt.NDArray, d: uint, dCurr: uint = 0) -> npt.NDArray:
                """
                Take derivative recursively.
                """
                if dCurr == d:
                    return dark
                else:
                    if dCurr == 0:
                        dark2 = np.hstack(
                            (Zero, 2.0 * One, np.zeros((N, self._m - 2), dtype=z.dtype))
                        )
                    else:
                        dark2 = np.zeros((N, self._m), dtype=z.dtype)
                    for k in range(1, self._m - 1):
                        dark2[:, k + 1 : k + 2] = (
                            2.0 * (dCurr + 1.0) * dark[:, k : k + 1]
                            + 2.0 * z * dark2[:, k : k + 1]
                            - 2.0 * k * dark2[:, k - 1 : k]
                        )
                    dCurr += 1
                    return Recurse(dark2, d, dCurr=dCurr)

            return Recurse(F, d)


class FS(BasisFunc):
    """
    Chebyshev polynomial basis functions.
    """

    def __init__(
        self,
        x0: Number,
        xf: Number,
        nC: npt.NDArray,
        m: uint,
    ) -> None:
        """
        Initialize the basis class.

        Parameters
        ----------
        x0 : Number
            Start of the problem domain.
        xf : Number
            End of the problem domain.
        nC : npt.NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        """
        super().__init__(x0, xf, nC, m, -np.pi, np.pi)

    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the CP basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """
        N = np.size(z)
        F = np.zeros((N, self._m))
        if d == 0:
            F[:, 0] = 1.0
            for k in range(1, self._m):
                g = np.ceil(k / 2.0)
                if k % 2 == 0:
                    F[:, k : k + 1] = np.cos(g * z)
                else:
                    F[:, k : k + 1] = np.sin(g * z)
        else:
            F[:, 0] = 0.0
            if d % 4 == 0:
                for k in range(1, self._m):
                    g = np.ceil(k / 2.0)
                    if k % 2 == 0:
                        F[:, k : k + 1] = g**d * np.cos(g * z)
                    else:
                        F[:, k : k + 1] = g**d * np.sin(g * z)
            elif d % 4 == 1:
                for k in range(1, self._m):
                    g = np.ceil(k / 2.0)
                    if k % 2 == 0:
                        F[:, k : k + 1] = -(g**d) * np.sin(g * z)
                    else:
                        F[:, k : k + 1] = g**d * np.cos(g * z)
            elif d % 4 == 2:
                for k in range(1, self._m):
                    g = np.ceil(k / 2.0)
                    if k % 2 == 0:
                        F[:, k : k + 1] = -(g**d) * np.cos(g * z)
                    else:
                        F[:, k : k + 1] = -(g**d) * np.sin(g * z)
            else:
                for k in range(1, self._m):
                    g = np.ceil(k / 2.0)
                    if k % 2 == 0:
                        F[:, k : k + 1] = g**d * np.sin(g * z)
                    else:
                        F[:, k : k + 1] = -(g**d) * np.cos(g * z)
        return F


class ELM(BasisFunc):
    """
    Extreme learning machine abstract basis class.
    """

    def __init__(
        self,
        x0: Number,
        xf: Number,
        nC: npt.NDArray,
        m: uint,
    ) -> None:
        """
        Initialize the basis class.

        Parameters
        ----------
        x0 : Number
            Start of the problem domain.
        xf : Number
            End of the problem domain.
        nC : npt.NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        """
        super().__init__(x0, xf, nC, m, 0.0, 1.0)

        dtype = np.array(self._c).dtype
        one = np.ones(1, dtype=dtype)

        self._w = np.random.uniform(low=-10.0, high=10.0, size=self._m) * one
        self._w = self._w.reshape((1, self._m))
        self._b = np.random.uniform(low=-10.0, high=10.0, size=self._m) * one
        self._b = self._b.reshape((1, self._m))

    @property
    def w(self) -> npt.NDArray:
        """
        Weights of the ELM

        Returns
        -------
        NDArray
            Weights of the ELM.
        """
        return self._w

    @property
    def b(self) -> npt.NDArray:
        """
        Biases of the ELM

        Returns
        -------
        NDArray
            Biases of the ELM.
        """
        return self._b

    @w.setter
    def w(self, val: npt.NDArray) -> None:
        """
        Weights of the ELM.

        Parameters
        ----------
        val : NDArray
            New weights.
        """
        if val.size == self._m:
            self._w = val
            if self._w.shape != (1, self._m):
                self._w = self._w.reshape((1, self._m))
        else:
            raise ValueError(
                f"Input array of size {val.size} was received, but size {self._m} was expected."
            )

    @b.setter
    def b(self, val: npt.NDArray) -> None:
        """
        Biases of the ELM.

        Parameters
        ----------
        val : NDArray
            New biases.
        """
        if val.size == self._m:
            self._b = val
            if self._b.shape != (1, self._m):
                self._b = self._b.reshape((1, self._m))
        else:
            raise ValueError(
                f"Input array of size {val.size} was received, but size {self._m} was expected."
            )


class ELMReLU(ELM):
    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the ELMRelu basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """

        if d == 0:
            return np.maximum(0.0, self._w * z + self._b)
        elif d == 1:
            return self._w * np.where(self._w * z + self._b > 0.0, 1.0, 0.0)
        else:
            return np.zeros((self._m, z.size))


class ELMSigmoid(ELM):
    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the ELMSigmoid basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """

        from tfc.utils import egrad

        f = lambda x: 1.0 / (1.0 + jnp.exp(-self._w * x - self._b))

        def Recurse(
            dark: Callable[[npt.NDArray], jnp.ndarray], d: uint, dCurr: uint = 0
        ) -> Callable[[npt.NDArray], jnp.ndarray]:
            """
            Take derivative recursively.
            """
            if dCurr == d:
                return dark
            else:
                dark2 = egrad(dark)
                dCurr += 1
                return Recurse(dark2, d, dCurr=dCurr)

        return np.asarray(Recurse(f, d)(z))


class ELMTanh(ELM):
    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the ELMTanh basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """

        from tfc.utils import egrad

        f = lambda x: jnp.tanh(self._w * x + self._b)

        def Recurse(
            dark: Callable[[npt.NDArray], jnp.ndarray], d: uint, dCurr: uint = 0
        ) -> Callable[[npt.NDArray], jnp.ndarray]:
            """
            Take derivative recursively.
            """
            if dCurr == d:
                return dark
            else:
                dark2 = egrad(dark)
                dCurr += 1
                return Recurse(dark2, d, dCurr=dCurr)

        return np.asarray(Recurse(f, d)(z))


class ELMSin(ELM):
    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the ELMSin basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """

        from tfc.utils import egrad

        f = lambda x: jnp.sin(self._w * x + self._b)

        def Recurse(
            dark: Callable[[npt.NDArray], jnp.ndarray], d: uint, dCurr: uint = 0
        ) -> Callable[[npt.NDArray], jnp.ndarray]:
            """
            Take derivative recursively.
            """
            if dCurr == d:
                return dark
            else:
                dark2 = egrad(dark)
                dCurr += 1
                return Recurse(dark2, d, dCurr=dCurr)

        return np.asarray(Recurse(f, d)(z))


class ELMSwish(ELM):
    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Internal method used to calcualte the ELMSwish basis function values.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : uint
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """

        from tfc.utils import egrad

        f = lambda x: (self._w * x + self._b) / (1.0 + jnp.exp(-self._w * x - self._b))

        def Recurse(
            dark: Callable[[npt.NDArray], jnp.ndarray], d: uint, dCurr: uint = 0
        ) -> Callable[[npt.NDArray], jnp.ndarray]:
            """
            Take derivative recursively.
            """
            if dCurr == d:
                return dark
            else:
                dark2 = egrad(dark)
                dCurr += 1
                return Recurse(dark2, d, dCurr=dCurr)

        return np.asarray(Recurse(f, d)(z))


class nBasisFunc(BasisFunc):
    """
    Python implementation of the n-dimensional basis function classes.
    See the Python implementation of `BasisFunc` for details.
    """

    def __init__(
        self,
        x0: npt.NDArray,
        xf: npt.NDArray,
        nC: npt.NDArray,
        m: uint,
        z0: Number = 0.0,
        zf: Number = 0.0,
    ) -> None:
        """
        Initialize the basis class.

        Parameters
        ----------
        x0 : NDArray
            Start of the problem domain.
        xf : NDArray
            End of the problem domain.
        nC : NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        z0 : Number
            Start of the basis function domain.
        zf : Number
            End of the basis function domain.
        """

        self._m = m
        self._nC = nC
        self._dim = nC.shape[0]
        self._numC = nC.shape[1]

        self._z0 = z0
        self._zf = zf
        self._x0 = x0
        if self._x0.shape != (self._dim, 1):
            self._x0 = self._x0.reshape((self._dim, 1))
        if xf.shape != (self._dim, 1):
            xf = xf.reshape((self._dim, 1))
        self._c = (zf - z0) / (xf - self._x0)

        vec = np.zeros((self._dim, 1))
        self._numBasisFunc = self._NumBasisFunc(self._dim - 1, vec, full=False)
        self._numBasisFuncFull = self._NumBasisFunc(self._dim - 1, vec, full=True)

    def _NumBasisFunc(self, dim: int, vec: npt.NDArray, n: int = 0, full: bool = False) -> int:
        """
        Calculate the number of basis functions.

        Parameters
        ----------
        dim : int
            Number of dimensions.
        vec : NDArray
            Vector used to keep track of the order of the basis function.
        n : int, optional
            Count of the number of basis functions so far. (Default value = 0)
        full : bool, optional
            If true, then does not remove basis functions based on self._nC. (Default value = False)

        Returns
        -------
        int
            Number of basis functions.
        """
        if dim > 0:
            for x in range(self._m):
                vec[dim] = x
                n = self._NumBasisFunc(dim - 1, vec, n=n, full=full)
        else:
            for x in range(self._m):
                vec[dim] = x
                if full:
                    if np.sum(vec) <= self._m - 1:
                        # If the degree of the produce of univariate basis functions is less than
                        # the degree specified, then add one to the count.
                        n += 1
                else:
                    if not np.all(np.any(vec == self._nC, axis=1)) and np.sum(vec) <= self._m - 1:
                        # If at least one of the dimensions' basis functions is not a constraint
                        # and the degree of the product of univariate basis functions is less than
                        # the degree specified, add one to the count
                        n += 1
        return n

    @property
    def c(self) -> npt.NDArray:
        """
        Return the constants that map the problem domain to the basis
        function domain.

        Returns
        -------
        npt.NDArray
            The constants that map the problem domain to the basis function
            domain.
        """

        return self._c

    @property
    def numBasisFunc(self) -> float:
        """
        Return the number of basis functions once user-specified
        functions have been removed.

        Returns
        -------
        float:
            The number of basis functions once the user-specified
            functions have been removed.
        """

        return self._numBasisFunc

    @property
    def numBasisFuncFull(self) -> float:
        """
        Return the number of basis functions before the user-specified
        functions have been removed.

        Returns
        -------
        float:
            The number of basis functions before the user-specified
            functions have been removed.
        """

        return self._numBasisFuncFull

    def H(self, x: npt.NDArray, d: npt.NDArray, full: bool = False) -> npt.NDArray:
        """
        Returns the basis function matrix for the x with a derivative of order d.

        Parameters
        -----------
        x : NDArray
            Input array. Values to calculate the basis function for.
            Should be size dim x N.
        d : NDArray
            Order of the derivative
        full: bool
            Whether to return the full basis function set, or remove
            the columns associated with self._nC.

        Returns
        -------
        H : NDArray
            The basis function values.
        """

        # Check dimensions
        N = x.shape[1]
        if x.shape[0] != self._dim:
            raise ValueError(
                f"Incorrect dimension for x. Expected {self._dim} but got {z.shape[1]}."
            )

        # Convert to basis function domain
        z = (x - self._x0) * self._c + self._z0

        # Create individual basis functions for each dimension
        T = np.zeros((N, self._m, self._dim), dtype=z.dtype)
        for k in range(self._dim):
            T[:, :, k] = self._Hint(z[k : k + 1, :].T, d[k]) * self._c[k] ** d[k]

        # Define functions for use in generating the CP sheet
        def MultT(vec: npt.NDArray) -> npt.NDArray:
            """
            Creates basis functions for the multidimensional case by mulitplying the basis functions
            for the single dimensional cases together.

            Parameters
            ----------
            vec : NDArray
                Used to track the basis functions used from the single dimensional cases.

            Returns
            -------
            NDArray
                Basis functions for the multidimensional case.
            """
            tout = np.ones((N, 1), dtype=z.dtype)
            for k in range(self._dim):
                tout *= T[:, vec[k, 0] : vec[k, 0] + 1, k]
            return tout

        def Recurse(
            dim: int, out: npt.NDArray, vec: npt.NDArray, n: int = 0, full: bool = False
        ) -> Tuple[npt.NDArray, int]:
            """
            Creates basis functions for the multidimensional case given the basis functions
            for the single dimensional cases.

            Parameters
            ----------
            dim : int
                Number of dimensions.
            out : NDArray
                Basis function for the multidimensional case created so far.
            n : int, optional
                Count of the number of basis functions created so far. (Default value = 0)
            full : bool, optional
                If true, then does not remove basis functions based on self._nC. (Default value = False)

            Returns
            -------
            out : NDArraY
                Basis functions for the multidimensional case created so far.
            n : int
                Basis function count.
            """
            if dim > 0:
                for x in range(self._m):
                    vec[dim] = x
                    out, n = Recurse(dim - 1, out, vec, n=n, full=full)
            else:
                for x in range(self._m):
                    vec[dim] = x
                    if full:
                        if np.sum(vec) <= self._m - 1:
                            # If the degree of the produce of univariate basis functions is less than
                            # the degree specified, then include this vector.
                            out[:, n : n + 1] = MultT(vec)
                            n += 1
                    else:
                        if (
                            not np.all(np.any(vec == self._nC, axis=1))
                            and np.sum(vec) <= self._m - 1
                        ):
                            # If at least one of the dimensions' basis functions is not a constraint
                            # and the degree of the product of univariate basis functions is less than
                            # the degree specified, include this vector.
                            out[:, n : n + 1] = MultT(vec)
                            n += 1
            return out, n

        # Calculate and store all possible combinations of the individual basis functions
        vec = np.zeros((self._dim, 1), dtype=int)
        if full:
            out = np.zeros((N, self._numBasisFuncFull), dtype=z.dtype)
        else:
            out = np.zeros((N, self._numBasisFunc), dtype=z.dtype)
        out, _ = Recurse(self._dim - 1, out, vec, full=full)

        return out


class nCP(nBasisFunc, CP):
    """
    n-dimensional Chebyshev polynomial basis functions.
    """

    def __init__(
        self,
        x0: npt.NDArray,
        xf: npt.NDArray,
        nC: npt.NDArray,
        m: uint,
    ) -> None:
        """
        Initialize the n-dimensional CP class.

        Parameters
        ----------
        x0 : NDArray
            Start of the problem domain.
        xf : NDArray
            End of the problem domain.
        nC : NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        """

        nBasisFunc.__init__(self, x0, xf, nC, m, -1.0, 1.0)


class nLeP(nBasisFunc, LeP):
    """
    n-dimensional Legendre polynomial basis functions.
    """

    def __init__(
        self,
        x0: npt.NDArray,
        xf: npt.NDArray,
        nC: npt.NDArray,
        m: uint,
    ) -> None:
        """
        Initialize the n-dimensional LeP class.

        Parameters
        ----------
        x0 : NDArray
            Start of the problem domain.
        xf : NDArray
            End of the problem domain.
        nC : NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        """

        nBasisFunc.__init__(self, x0, xf, nC, m, -1.0, 1.0)


class nFS(nBasisFunc, FS):
    """
    n-dimensional Fourier series basis functions.
    """

    def __init__(
        self,
        x0: npt.NDArray,
        xf: npt.NDArray,
        nC: npt.NDArray,
        m: uint,
    ) -> None:
        """
        Initialize the n-dimensional FS class.

        Parameters
        ----------
        x0 : NDArray
            Start of the problem domain.
        xf : NDArray
            End of the problem domain.
        nC : NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        """

        nBasisFunc.__init__(self, x0, xf, nC, m, -np.pi, np.pi)


class nELM(nBasisFunc):
    """
    n-dimensional extreme learning machine abstract basis class.
    """

    def __init__(
        self,
        x0: npt.NDArray,
        xf: npt.NDArray,
        nC: npt.NDArray,
        m: uint,
        z0: Number = 0.0,
        zf: Number = 1.0,
    ) -> None:
        """
        Initialize the basis class.

        Parameters
        ----------
        x0 : NDArray
            Start of the problem domain.
        xf : NDArray
            End of the problem domain.
        nC : NDArray
            Basis functions to be removed
        m : uint
            Number of basis functions.
        z0 : Number
            Start of the basis function domain.
        zf : Number
            End of the basis function domain.
        """

        self._m = m
        self._nC = nC
        self._dim = x0.size

        if np.any(self._nC != -1):
            self._numC = nC.size
        else:
            self._numC = 0

        self._z0 = z0
        self._zf = zf
        self._x0 = x0
        if self._x0.shape != (self._dim, 1):
            self._x0 = self._x0.reshape((self._dim, 1))
        if xf.shape != (self._dim, 1):
            xf = xf.reshape((self._dim, 1))
        self._c = (zf - z0) / (xf - self._x0)

        self._numBasisFunc = self._m - self._numC
        self._numBasisFuncFull = self._m

        one = np.ones(1, dtype=x0.dtype)
        self._w = np.random.uniform(low=-1.0, high=1.0, size=self._dim * self._m) * one
        self._w = self._w.reshape((self._dim, self._m))
        self._b = np.random.uniform(low=-1.0, high=1.0, size=self._m) * one
        self._b = self._b.reshape((1, self._m))

    @property
    def w(self) -> npt.NDArray:
        """
        Weights of the nELM

        Returns
        -------
        NDArray
            Weights of the ELM.
        """
        return self._w

    @property
    def b(self) -> npt.NDArray:
        """
        Biases of the nELM

        Returns
        -------
        NDArray
            Biases of the ELM.
        """
        return self._b

    @w.setter
    def w(self, val: npt.NDArray) -> None:
        """
        Weights of the nELM.

        Parameters
        ----------
        val : NDArray
            New weights.
        """
        if val.size == self._m * self._dim:
            self._w = val
            if self._w.shape != (self._dim, self._m):
                self._w = self._w.reshape((self._dim, self._m))
        else:
            raise ValueError(
                f"Input array of size {val.size} was received, but size {self._m*self._dim} was expected."
            )

    @b.setter
    def b(self, val: npt.NDArray) -> None:
        """
        Biases of the nELM.

        Parameters
        ----------
        val : NDArray
            New biases.
        """
        if val.size == self._m:
            self._b = val
            if self._b.shape != (1, self._m):
                self._b = self._b.reshape((1, self._m))
        else:
            raise ValueError(
                f"Input array of size {val.size} was received, but size {self._m} was expected."
            )

    def H(self, x: npt.NDArray, d: npt.NDArray, full: bool = False) -> npt.NDArray:
        """
        Returns the basis function matrix for the x with a derivative of order d.

        Parameters
        ----------
        x : NDArray
            Input array. Values to calculate the basis function for.
            Should be size dim x N.
        d : NDArray
            Order of the derivative
        full : bool
            Whether to return the full basis function set, or remove
            the columns associated with self._nC.

        Returns
        -------
        H : NDArray
            The basis function values.
        """

        # Check dimensions
        if x.shape[0] != self._dim:
            raise ValueError(
                f"Incorrect dimension for x. Expected {self._dim} but got {z.shape[1]}."
            )

        # Convert to basis function domain
        z = ((x - self._x0) * self._c + self._z0).T

        F = self._nHint(z, d)
        if not full and self._numC > 0:
            F = np.delete(F, self._nC, axis=1)
        return F

    @abstractmethod
    def _nHint(self, z: npt.NDArray, d: npt.NDArray) -> npt.NDArray:
        """
        Internal method used to calcualte the basis function value.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : NDArray
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """
        pass

    def _Hint(self, z: npt.NDArray, d: uint) -> npt.NDArray:
        """
        Dummy function, this should never be called!
        """
        raise ValueError("Error: This function should never be called.")


class nELMReLU(nELM):
    """
    n-dimensional ELM ReLU basis functions.
    """

    def _nHint(self, z: npt.NDArray, d: npt.NDArray) -> npt.NDArray:
        """
        Internal method used to calcualte the basis function value.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : NDArray
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """
        ind = -1
        zeroFlag = False
        dorder = np.sum(d)
        if dorder > 1:
            zeroFlag = True
        elif dorder == 1:
            ind = np.where(d == 1)[0][0]

        if zeroFlag:
            # Derivative order is high enough that everything is zeros
            return np.zeros((z.shape[0], self._m))
        elif ind != -1:
            # We have a derivative on only one variable
            return (
                self._c[ind]
                * self._w[ind : ind + 1, :]
                * np.where(np.dot(z, self._w) + self._b > 0.0, 1.0, 0.0)
            )
        else:
            return np.maximum(0.0, np.dot(z, self._w) + self._b)


class nELMSin(nELM):
    """
    n-dimensional ELM sin basis functions.
    """

    def _nHint(self, z: npt.NDArray, d: npt.NDArray) -> npt.NDArray:
        """
        Internal method used to calcualte the basis function value.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : NDArray
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """

        from tfc.utils import egrad

        f = lambda *x: jnp.sin(jnp.dot(jnp.hstack(x), self._w) + self._b)

        z = jnp.split(z, z.shape[1], axis=1)

        def Recurse(
            dark: Callable[[npt.NDArray], jnp.ndarray], d: uint, dim: uint, dCurr: uint = 0
        ) -> Callable[[npt.NDArray], jnp.ndarray]:
            if dCurr == d:
                return dark
            else:
                dark2 = egrad(dark, dim)
                dCurr += 1
                return Recurse(dark2, d, dim, dCurr=dCurr)

        dark = f
        dark2 = 1
        for dim, deriv in enumerate(d):
            dark2 *= self._c[dim] ** deriv
            dark = Recurse(dark, deriv, dim)

        return np.asarray((dark(*z) * dark2))


class nELMTanh(nELM):
    """
    n-dimensional ELM tanh basis functions.
    """

    def _nHint(self, z: npt.NDArray, d: npt.NDArray) -> npt.NDArray:
        """
        Internal method used to calcualte the basis function value.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : NDArray
            Derivative order.

        Returns
        -------
        H: NDArray
            Basis function values.
        """

        from tfc.utils import egrad

        f = lambda *x: jnp.tanh(jnp.dot(jnp.hstack(x), self._w) + self._b)

        z = jnp.split(z, z.shape[1], axis=1)

        def Recurse(
            dark: Callable[[npt.NDArray], jnp.ndarray], d: uint, dim: uint, dCurr: uint = 0
        ) -> Callable[[npt.NDArray], jnp.ndarray]:
            if dCurr == d:
                return dark
            else:
                dark2 = egrad(dark, dim)
                dCurr += 1
                return Recurse(dark2, d, dim, dCurr=dCurr)

        dark = f
        dark2 = 1
        for dim, deriv in enumerate(d):
            dark2 *= self._c[dim] ** deriv
            dark = Recurse(dark, deriv, dim)

        return np.asarray((dark(*z) * dark2))


class nELMSigmoid(nELM):
    """
    n-dimensional ELM sigmoid basis functions.
    """

    def _nHint(self, z: npt.NDArray, d: npt.NDArray) -> npt.NDArray:
        """
        Internal method used to calcualte the basis function value.

        Parameters:
        -----------
        z: NDArray
            Values to calculate the basis functions for.
        d: NDArray
            Derivative order.

        Returns:
        --------
        H: NDArray
            Basis function values.
        """

        from tfc.utils import egrad

        f = lambda *x: 1.0 / (1.0 + jnp.exp(-jnp.dot(jnp.hstack(x), self._w) - self._b))

        z = jnp.split(z, z.shape[1], axis=1)

        def Recurse(
            dark: Callable[[npt.NDArray], jnp.ndarray], d: uint, dim: uint, dCurr: uint = 0
        ) -> Callable[[npt.NDArray], jnp.ndarray]:
            if dCurr == d:
                return dark
            else:
                dark2 = egrad(dark, dim)
                dCurr += 1
                return Recurse(dark2, d, dim, dCurr=dCurr)

        dark = f
        dark2 = 1
        for dim, deriv in enumerate(d):
            dark2 *= self._c[dim] ** deriv
            dark = Recurse(dark, deriv, dim)

        return np.asarray((dark(*z) * dark2))


class nELMSwish(nELM):
    """
    n-dimensional ELM swish basis functions.
    """

    def _nHint(self, z: npt.NDArray, d: npt.NDArray) -> npt.NDArray:
        """
        Internal method used to calcualte the basis function value.

        Parameters
        ----------
        z : NDArray
            Values to calculate the basis functions for.
        d : NDArray
            Derivative order.

        Returns
        -------
        H : NDArray
            Basis function values.
        """

        from tfc.utils import egrad

        def f(*x):
            dark = jnp.dot(jnp.hstack(x), self._w) + self._b
            return dark / (1.0 + jnp.exp(-dark))

        z = jnp.split(z, z.shape[1], axis=1)

        def Recurse(
            dark: Callable[[npt.NDArray], jnp.ndarray], d: uint, dim: uint, dCurr: uint = 0
        ) -> Callable[[npt.NDArray], jnp.ndarray]:
            if dCurr == d:
                return dark
            else:
                dark2 = egrad(dark, dim)
                dCurr += 1
                return Recurse(dark2, d, dim, dCurr=dCurr)

        dark = f
        dark2 = 1
        for dim, deriv in enumerate(d):
            dark2 *= self._c[dim] ** deriv
            dark = Recurse(dark, deriv, dim)

        return np.asarray((dark(*z) * dark2))
